#!/usr/bin/env python3
import logging
import os
from pyparsing import ParseException
from quit.exceptions import UnSupportedQuery, SparqlProtocolError, NonAbsoluteBaseError
from rdflib.term import URIRef
from rdflib.plugins.sparql.parserutils import CompValue, plist
from rdflib.plugins.sparql.parser import parseQuery, parseUpdate
from quit.tools.algebra import translateQuery, translateUpdate
from rdflib.plugins.serializers.nt import _nt_row as _nt
from rdflib.plugins.sparql import parser, algebra
from rdflib.plugins import sparql
from uritools import urisplit
from werkzeug.http import parse_options_header

logger = logging.getLogger('quit.helpers')


class QueryAnalyzer:
    """A class that provides methods for received sparql query strings.

    This class is used to classify a given query string.
    At the moment the class distinguishes between SPARQL Update and Select queries.
    """

    logger = logging.getLogger('quit.helpers.QueryAnalyzer')

    def __init__(self, querystring, graph=None):
        """Initialize a check for a given query string.

        Args:
            querystring: A string containing a query.
        """
        self.query = querystring
        self.parsedQuery = None
        self.queryType = None
        self.actions = None

        if self.evalQuery(querystring):
            return

        if self.evalUpdate(querystring, graph):
            return

        return

    def prepareUpdate(self, updateString, initNs={}, base=None):
        """Parse and translate a SPARQL Query."""
        parsedUpdate = parser.parseUpdate(str(updateString))
        return algebra.translateUpdate(parsedUpdate, base, initNs)

    def getType(self):
        """Return the type of a query.

        Returns:
            A string containing the query type.
        """
        return self.queryType

    def getActions(self):
        """Return the type of a query.

        Returns:
            A string containing the query type.
        """
        return self.actionsType

    def getParsedQuery(self):
        """Return the query object (rdflib) of a query string.

        Returns:
            The query object after a query string was parsed with Rdflib.
        """
        return self.parsedQuery

    def evalQuery(self, querystring):
        """Check if the given querystring contains valid SPARQL queries.

        Returns:
            True, if querystring is valid.
            Else, if not.
        """
        try:
            self.parsedQuery = sparql.prepareQuery(querystring)
            logger.debug(str(self.parsedQuery.algebra.name))
            if str(self.parsedQuery.algebra.name) == 'DescribeQuery':
                self.queryType = 'DESCRIBE'
            elif str(self.parsedQuery.algebra.name) == 'ConstructQuery':
                self.queryType = 'CONSTRUCT'
            elif str(self.parsedQuery.algebra.name) == 'SelectQuery':
                self.queryType = 'SELECT'
            elif str(self.parsedQuery.algebra.name) == 'AskQuery':
                self.queryType = 'ASK'
            return True
        except Exception:
            return False

    def evalUpdate(self, querystring, graph):
        """Check if the given querystring contains (a) valid SPARQL update query(ies).

        Returns:
            True, if querystring is valid.
            Else, if not.
        """
        self.parsedQuery = self.prepareUpdate(querystring)
        self.queryType = 'UPDATE'
        return


def applyChangeset(f, changeset, identifier):
    """Update the FileReference (graph uri) of a file with help of the changeset."""
    for (op, triples) in changeset:
        if op == 'additions':
            for triple in triples:
                # the internal _nt serializer appends '\n'
                line = _nt(triple).rstrip()
                f.add(line)
        elif op == 'removals':
            for triple in triples:
                # the internal _nt serializer appends '\n'
                line = _nt(triple).rstrip()
                f.remove(line)


def isAbsoluteUri(uri):
    """Check if a URI is a absolute URI and uses 'http(s)' at protocol part.

    Returns:
        True, if absolute http(s) URIs
        False, if not
    """
    try:
        parsed = urisplit(uri)
    except Exception:
        return False
    # We accept Absolute URI as specified in https://tools.ietf.org/html/rfc3986#section-4.3
    # with http(s) scheme
    if parsed[0] and parsed[0] in ['http', 'https'] and parsed[1] and not parsed[4] and (
            parsed[2] == '/' or os.path.isabs(parsed[2])):
        return True
    else:
        return False


def configure_query_dataset(parsed_query, default_graphs, named_graphs):
    """Substitute the default and named graph URI.

    According to https://www.w3.org/TR/sparql11-protocol/ we will remove the named and default graph
    URIs given in the query string (if given) and will add default-graph-uri and named-graph-uri
    from protocol request.

    Args: parsed_query: the parsed query
          default_graphs: a list of uri strings for default graphs
          named_graphs: a list of uri strings for named graphs
    """
    if not isinstance(default_graphs, list) or not isinstance(named_graphs, list):
        return parsed_query

    if len(default_graphs) == 0 and len(named_graphs) == 0:
        return parsed_query

    # clean existing named (FROM NAMED) and default (FROM) DatasetClauses
    parsed_query[1]['datasetClause'] = plist()

    # add new named (default-graph-uri) and default (named-graph-uri)
    # DatasetClauses from Protocol
    for uri in default_graphs:
        parsed_query[1]['datasetClause'].append(CompValue('DatasetClause', default=URIRef(uri)))
    for uri in named_graphs:
        parsed_query[1]['datasetClause'].append(CompValue('DatasetClause', named=URIRef(uri)))

    return parsed_query


def configure_update_dataset(parsed_update, default_graphs, named_graphs):
    """Add default and named graph URI.

    According to https://www.w3.org/TR/sparql11-protocol/ we will add using-named-graph-uri and
    using-graph-uri if the update requst does not contain a USING, USING NAMED, or WITH clause.

    Args: parsed_update: the parsed update
          default_graphs: a list of uri strings for default graphs
          named_graphs: a list of uri strings for named graphs
    """
    if not isinstance(default_graphs, list) or not isinstance(named_graphs, list):
        return parsed_update

    if len(default_graphs) == 0 and len(named_graphs) == 0:
        return parsed_update

    if parsed_update.request[0].withClause is not None:
        raise SparqlProtocolError

    if parsed_update.request[0].using is not None:
        raise SparqlProtocolError

    parsed_update.request[0]['using'] = plist()

    # add new named (using-named-graph-uri) and default (using-graph-uri)
    # UsingClauses from Protocol
    for uri in default_graphs:
        parsed_update.request[0]['using'].append(CompValue('UsingClause', default=URIRef(uri)))
    for uri in named_graphs:
        parsed_update.request[0]['using'].append(CompValue('UsingClause', named=URIRef(uri)))

    return parsed_update


def parse_query_type(query, base=None, default_graph=[], named_graph=[]):
    """Parse a query and add default and named graph uri if possible."""
    try:
        parsed_query = parseQuery(query)
        parsed_query = configure_query_dataset(parsed_query, default_graph, named_graph)
        translated_query = translateQuery(parsed_query, base=base)
    except ParseException:
        raise UnSupportedQuery()
    except SparqlProtocolError as e:
        raise e

    if base is not None and not isAbsoluteUri(base):
        raise NonAbsoluteBaseError()

    if not is_valid_query_base(parsed_query):
        raise NonAbsoluteBaseError()

    return translated_query.algebra.name, translated_query


def parse_update_type(query, base=None, default_graph=[], named_graph=[]):
    """Parse an update and add default and named graph uri if possible."""
    try:
        parsed_update = parseUpdate(query)
        parsed_update = configure_update_dataset(parsed_update, default_graph, named_graph)
        translated_update = translateUpdate(parsed_update, base=base)
    except ParseException:
        raise UnSupportedQuery()
    except SparqlProtocolError as e:
        raise e

    if base is not None and not isAbsoluteUri(base):
        raise NonAbsoluteBaseError()

    if not is_valid_update_base(parsed_update):
        raise NonAbsoluteBaseError()

    return parsed_update.request[0].name, translated_update


def is_valid_query_base(parsed_query):
    """Check if a query contains an absolute base if base is given.

    Args: parsed_query: the parsed query
    Returns: True - if Base URI is given and abolute or if no Base is given
             False - if Base URI is given an not absolute
    """
    for value in parsed_query[0]:
        if value.name == 'Base' and not isAbsoluteUri(value.iri):
            return False

    return True


def is_valid_update_base(parsed_update):
    """Check if an update contains an absolute base if base is given.

    Args: parsed_update: the parsed update
    Returns: True - if Base URI is given and abolute or if no Base is given
             False - if Base URI is given an not absolute
    """
    for value in parsed_update.prologue[0]:
        if value.name == 'Base' and not isAbsoluteUri(value.iri):
            return False

    return True


def parse_sparql_request(request):
    """Parse a request according to SPARQL 1.1. protocol and return needed information.

    Args:
        request: A flask HTTP request
    Returns:
        quintuple - query, type, mimetype, default_graph, named_graph
    """
    query = None
    type = None
    default_graph = []
    named_graph = []
    accept_header = None

    if request.method == "GET":
        default_graph = request.args.getlist('default-graph-uri')
        named_graph = request.args.getlist('named-graph-uri')
        query = request.args.get('query', None)
        type = 'query'
    elif request.method == "POST":
        if 'Content-Type' in request.headers:
            content_mimetype, options = parse_options_header(request.headers['Content-Type'])
            if content_mimetype == "application/x-www-form-urlencoded":
                if 'query' in request.form:
                    default_graph = request.form.getlist('default-graph-uri')
                    named_graph = request.form.getlist('named-graph-uri')
                    query = request.form.get('query', None)
                    type = 'query'
                elif 'update' in request.form:
                    default_graph = request.form.getlist('using-graph-uri')
                    named_graph = request.form.getlist('using-named-graph-uri')
                    query = request.form.get('update', None)
                    type = 'update'
            elif content_mimetype == "application/sparql-query":
                default_graph = request.args.getlist('default-graph-uri')
                named_graph = request.args.getlist('named-graph-uri')
                query = request.data.decode("utf-8")
                type = 'query'
            elif content_mimetype == "application/sparql-update":
                default_graph = request.args.getlist('using-graph-uri')
                named_graph = request.args.getlist('using-named-graph-uri')
                query = request.data.decode("utf-8")
                type = 'update'

    return query, type, default_graph, named_graph



"""
Generate preformatted data of commits graph.

https://github.com/tclh123/commits-graph

Copyright (c) 2014
MIT License
"""

import json

from itertools import islice
from pygit2 import (
    GIT_SORT_TIME,
    GIT_SORT_TOPOLOGICAL,
    GIT_SORT_REVERSE
)


class CommitGraph(object):
    def __init__(self, sha, parents):
        self.sha = sha
        self.parents = parents

    @classmethod
    def gets(cls, commits):
        return [cls(c.id, [p.id for p in c.parents]) for c in commits]


def generate_graph_data(commits):
    """Generate graph data.

    :param commits: a list of commit, which should have
        `sha`, `parents` properties.
    :returns: data nodes, a json list of
        [
          sha,
          [offset, branch], //dot
          [
            [from, to, branch],  // route 1
            [from, to, branch],  // route 2
            [from, to, branch],
          ]  // routes
        ],  // node
    """

    nodes = []
    branch_idx = [0]
    reserve = []
    branches = {}

    def get_branch(sha):
        if branches.get(sha) is None:
            branches[sha] = branch_idx[0]
            reserve.append(branch_idx[0])
            branch_idx[0] += 1
        return branches[sha]

    for commit in commits:
        branch = get_branch(commit.sha)
        n_parents = len(commit.parents)
        offset = reserve.index(branch)
        routes = []

        if n_parents == 1:
            # create branch
            if branches.get(commit.parents[0]) is not None:
                routes += [[i + offset + 1, i + offset + 1 - 1, b]
                           for i, b in enumerate(reserve[offset + 1:])]
                routes += [[i, i, b]
                           for i, b in enumerate(reserve[:offset])]
                reserve.remove(branch)
                parent_branch = branches[commit.parents[0]]
                if parent_branch in reserve:
                    reserve_branch = reserve.index(parent_branch)
                    routes.append([offset, reserve_branch, branch])
            # straight
            else:
                routes += [[i, i, b] for i, b in enumerate(reserve)]
                branches[commit.parents[0]] = branch
        # merge branch
        elif n_parents == 2:
            branches[commit.parents[0]] = branch
            routes += [[i, i, b] for i, b in enumerate(reserve)]
            other_branch = get_branch(commit.parents[1])
            if other_branch in reserve:
                routes.append([offset, reserve.index(other_branch),
                               other_branch])

        node = _make_node(commit.sha,
                          offset,
                          branch,
                          routes)
        nodes.append(node)

    return json.dumps(nodes)


def _make_node(sha, offset, branch, routes):
    return [sha, [offset, branch], routes]



import os
import pygit2
import re
import logging

from _pygit2 import GitError, Oid
from os.path import expanduser, join
from quit.exceptions import RepositoryNotFound, RevisionNotFound, NodeNotFound, RemoteNotFound
from quit.exceptions import QuitGitRefNotFound, QuitGitRepoError, QuitGitPushError
from quit.namespace import QUIT
from quit.merge import Merger

import subprocess

logger = logging.getLogger('quit.git')

# roles
role_author = QUIT['author']
role_committer = QUIT['committer']


class Repository(object):
    """The Quit class for wrapping a git repository.

    TODO:
    - There is no possibility to set remotes on a Quit Repository object.
    """

    def __init__(self, path, origin=None, create=False, garbageCollection=False, callback=None):
        """Initialize a quit repo at a given location of the filesystem.

        Keyword arguments:
        path -- the location of an existing git repository or where the repository should be created
        origin -- if the repository does not exist and origin is given the origin repository is
                  cloned to the given path (default: None)
        create -- boolean whether to create a new repository if the specified path is empty
                  (default: False)
        garbageCollection -- boolean whether to activate the garbage collection on the git
                  repository (default: False)
        callback -- an instance of pygit2.RemoteCallbacks to handle cedentials and
                  push_update_reference (default: None)
        """
        if not callback:
            callback = QuitRemoteCallbacks()
        self.callback = callback

        try:
            self._repository = pygit2.Repository(path)
        except (KeyError, GitError):
            if not create:
                raise RepositoryNotFound(
                    'Repository "%s" does not exist' % path)

            if origin:
                self._repository = pygit2.clone_repository(
                    url=origin, path=path, bare=False, callbacks=self.callback
                )
            else:
                self._repository = pygit2.init_repository(path)

        if garbageCollection:
            self.init_garbageCollection(path)

        self.path = path

    def init_garbageCollection(self, path):
        """Set the threshold for automatic garbage collection for the git repository."""
        try:
            with subprocess.Popen(
                ["git", "config", "gc.auto"],
                stdout=subprocess.PIPE,
                cwd=path
            ) as gcAutoThresholdProcess:
                stdout, stderr = gcAutoThresholdProcess.communicate()
                gcAutoThreshold = stdout.decode("UTF-8").strip()

            if not gcAutoThreshold:
                gcAutoThreshold = 256
                subprocess.Popen(["git", "config", "gc.auto", str(gcAutoThreshold)], cwd=path)
                logger.info("Set default gc.auto threshold {}".format(gcAutoThreshold))

            logger.info(
                "Garbage Collection is enabled with gc.auto threshold {}".format(gcAutoThreshold)
            )
        except Exception as e:
            # Disable garbage collection for the rest of the run because it
            # is likely that git is not available
            logger.info('Git garbage collection could not be configured and was disabled')
            logger.debug(e)

    @property
    def is_empty(self):
        return self._repository.is_empty

    @property
    def is_bare(self):
        return self._repository.is_bare

    def close(self):
        self._repository = None

    def lookup(self, name):
        """Lookup the oid for a reference.

        The name is looked for in "refs/heads/<name>", "refs/tags/<name>" and directly.
        It does not matter weather the found reference is a symbolic or a direct, it will be
        resolved to an oid.

        Return:
        Oid
        """
        try:
            for template in ['refs/heads/%s', 'refs/tags/%s', '%s']:
                try:
                    reference = self._repository.lookup_reference(template % name)
                    return reference.resolve().target
                except KeyError:
                    pass
        except ValueError:
            pass
        try:
            revison = self._repository.get(name)
            if revison:
                return revison.id
        except Exception as e:
            logger.exception(e)
        raise RevisionNotFound(name)

    def revision(self, id='HEAD'):
        try:
            commit = self._repository.revparse_single(id)
        except KeyError:
            raise RevisionNotFound(id)

        return Revision(self, commit)

    def revisions(self, name=None, order=pygit2.GIT_SORT_REVERSE):
        seen = set()

        def traverse(ref, seen):
            for commit in self._repository.walk(ref, order):
                oid = commit.oid
                if oid not in seen:
                    seen.add(oid)
                    yield Revision(self, commit)

        def iter_commits(name, seen):
            commits = []

            if not name:
                for name in self.branches:
                    ref = self.lookup(name)
                    commits += traverse(ref, seen)
            else:
                oid = self.lookup(name)
                commits += traverse(oid, seen)
            return commits

        return iter_commits(name, seen)

    @property
    def current_head(self):
        if not self._repository.head_is_unborn:
            return re.sub("refs/heads/", "", self._repository.head.name)
        return None

    @property
    def branches(self):
        """Get a list of all branch names."""
        return [x[len('refs/heads/'):] for x in self._repository.listall_references()
                if x.startswith('refs/heads/')]

    @property
    def tags(self):
        """Get a list of all tag names."""
        return [x[len('refs/tags/'):] for x in self._repository.listall_references()
                if x.startswith('refs/tags/')]

    @property
    def references(self):
        """Get a list of all references."""
        return self._repository.listall_references()

    @property
    def remotes(self):
        """Get a list of all configured remotes."""
        return [{"name": x.name, "url": x.url} for x in self._repository.remotes]

    @property
    def tags_or_branches(self):
        """Get a list of all tag and head references."""
        return [
            x for x in self._repository.listall_references()
            if x.startswith('refs/tags/') or x.startswith('refs/heads/')
        ]

    def index(self, revision=None):
        index = Index(self)
        if revision:
            index.set_revision(revision)
        return index

    def getUpstreamOfHead(self):
        if self._repository.head_is_unborn:
            return (None, None, None)
        localBranchName = re.sub("^refs/heads/", "", self._repository.head.name)
        localBranch = self._repository.lookup_branch(localBranchName)
        if localBranch.upstream is not None:
            refspecPattern = re.compile("^refs/remotes/([^/]*)/(.*)$")
            remote_name = localBranch.upstream.remote_name
            remote_ref = localBranch.upstream.name
            if remote_ref.startswith("refs/remotes/"):
                matchgroups = refspecPattern.match(remote_ref)
                remote_branch = matchgroups.group(2)
            else:
                remote_branch = remote_ref
            return (remote_name, remote_branch, remote_ref)
        return (None, None, None)

    def fetch(self, remote_name=None, remote_branch=None):
        """Fetch changes from a remote.

        If no remote_name and no remote_branch is given, the method will fetch changes from the
        remote called origin or if the current branch referenced by HEAD has upstream configured, it
        will fetch the configured upstream branch.

        Keyword arguments:
        remote_name --The name of the remote from where to fetch.
        remote_branch -- The name of the remote branch to fetch or a local reference to a remote
                         branch (refs/remotes/.../...)
        """
        remote_ref = None
        if (remote_name is None and
                remote_branch is not None and
                remote_branch.startswith("refs/remotes/")):
            refspecPattern = re.compile("^refs/remotes/([^/]*)/(.*)$")
            remote_ref = remote_branch
            matchgroups = refspecPattern.match(remote_branch)
            remote_name = matchgroups.group(1)
            remote_branch = matchgroups.group(2)
        if remote_branch is None and not self._repository.head_is_unborn:
            (head_remote_name, head_remote_branch, head_remote_ref) = self.getUpstreamOfHead()
            if remote_name is None or remote_name == head_remote_name:
                remote_name = head_remote_name
                remote_branch = head_remote_branch
                remote_ref = head_remote_ref
        if remote_name is None:
            remote_name = 'origin'
        else:
            remote_ref = 'refs/remotes/{remote}/{remote_branch}'.format(remote=remote_name,
                                                                        remote_branch=remote_branch)
        logger.debug('Fetching: {} from {} ({})'.format(remote_branch or "all references",
                                                        remote_name, remote_ref))
        for remote in self._repository.remotes:
            if remote.name == remote_name:
                if remote_branch is not None:
                    remote_branch = [remote_branch]
                remote.fetch(remote_branch, callbacks=self.callback)

                if remote_branch is None:
                    return None

                return self._repository.lookup_reference(remote_ref).target
        raise RemoteNotFound("There is no remote \"{}\".".format(remote_name))

    def pull(self, remote_name=None, refspec=None, **kwargs):
        """Pull (fetch and merge) changes from a remote repository.

        Keyword arguments:
        remote_name -- The name of a remote repository as configured on the underlaying git
                       repository (default: "origin")
        refspec -- The refspec of remote branch to fetch and the local reference/branch to update.
                   An optional plus (+) in the beginning of the refspec is ignored.
        """
        remote_branch = None
        local_branch = None
        if refspec is not None:
            rerefspec = re.compile("^(?P<plus>[+]?)(?P<src>[^:]*)(:(?P<dst>.*))?$")
            groups = rerefspec.match(refspec)
            if groups is None:
                raise QuitGitRepoError("The respec \"{}\" could not be understood".format(refspec))
            plus = groups.group("plus")
            remote_branch = groups.group("src")
            local_branch = groups.group("dst")
            logger.debug("pull: parsed refspec is: {}, {}, {}".format(plus, remote_branch,
                                                                      local_branch))
        remote_reference = self.fetch(remote_name=remote_name, remote_branch=remote_branch)
        if remote_reference is not None:
            self.merge(target=local_branch, branch=remote_reference, **kwargs)

    def push(self, remote_name=None, refspec=None):
        """Push changes on a local repository to a remote repository.

        Keyword arguments:
        remote_name -- The name of a remote repository as configured on the underlaying git
                       repository (default: "origin")
        refspec -- The local and remote reference to push divided with a colon.
                   (refs/heads/master:refs/heads/master)
        """
        if refspec is None:
            (head_remote_name, head_remote_branch, head_remote_ref) = self.getUpstreamOfHead()
            if head_remote_ref is None:
                raise QuitGitRefNotFound("There is no upstream configured for the current branch")
            refspec = '{src}:{dst}'.format(self._repository.head.name, head_remote_ref)
            remote_name = head_remote_name

        if remote_name is None:
            remote_name = 'origin'

        try:
            left, right = refspec.split(':')
        except Exception:
            left = refspec
            right = None

        refspec = ""
        if left:
            if not left.startswith("refs/heads/"):
                refspec = "refs/heads/" + left
            else:
                refspec = left
        if right:
            if not right.startswith('refs/heads/'):
                refspec += ':refs/heads/' + right
            else:
                refspec += ':' + right

        logger.debug("push: refspec {} to {}".format(refspec, remote_name))

        for remote in self._repository.remotes:
            if remote.name == remote_name:
                remote.push([refspec], callbacks=self.callback)
                if self.callback.push_error is not None:
                    raise self.callback.push_error
                return
        raise RemoteNotFound("There is no remote \"{}\".".format(remote_name))

    def merge(self, target=None, branch=None, method=None):
        """Merge a branch into another (target) branch.

        Merge 'branch' into 'target' and set 'target' to the resulting commit.

        Keyword arguments:
        target -- The target of the merge operation (if omitted, 'branch' will be merged into HEAD)
        branch -- A string with the branche name which should be merged into 'target' or an Oid of a
                  commit to be merged (the type has to be pygit2.Oid). (defaults to FEATCH_HEAD)
        """
        logger.debug("Start Merge")
        if target is None:
            target = "HEAD"
        if branch is None:
            branch = "FETCH_HEAD"
        if method is None:
            method = "three-way"

        if target not in ["HEAD", "FETCH_HEAD"] and not target.startswith("refs/heads/"):
            target = "refs/heads/" + target

        if isinstance(branch, pygit2.Oid):
            branch = str(branch)
        elif branch not in ["HEAD", "FETCH_HEAD"] and not branch.startswith("refs/heads/"):
            branch = "refs/heads/" + branch

        logger.debug("merge: {} into {} with {}".format(branch, target, method))

        merger = Merger(self, self._repository)

        merge_result = merger.merge_analysis(target, branch)

        # Up to date, do nothing
        if merge_result & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
            logger.debug("merge {} into {} everything up to date".format(branch, target))
            return pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE

        target_ref = self._repository.lookup_reference(target)

        # We can just fastforward
        if merge_result & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
            logger.debug("merge {} into {} we are doing a FFW".format(branch, target))
            try:
                if target_ref.type == pygit2.GIT_REF_SYMBOLIC:
                    target_ref = target_ref.resolve()
                target_ref.set_target(self.lookup(branch))
            except KeyError as e:
                # If HEAD is unborn
                if target_ref.type == pygit2.GIT_REF_SYMBOLIC:
                    target = target_ref.target
                target_ref = self._repository.create_reference(target, self.lookup(branch))

            return pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD

        if merge_result & pygit2.GIT_MERGE_ANALYSIS_NORMAL:

            if ((target == "HEAD" or target == self._repository.head.name) and
                    method == "three-way-git"):
                logger.debug("merge {} into {} three-way-git with HEAD".format(branch, target))
                merger.merge_three_way_head(branch)
                return merge_result
            elif method in ["three-way", "context"]:
                logger.debug("merge {} into {} three-way or context for RDF".format(branch, target))
                merger.merge_quit_commits(target, branch, favour=method)
                return merge_result

            logger.debug("merge {} into {} with {} is not yet supported".format(branch, target,
                                                                                method))
            raise Exception("Not yet supported merge method")

        raise AssertionError('Unknown merge analysis result')

    def branch(self, oldbranch=None, newbranch=None):
        """Create a new branch from an existing branch."""
        logger.debug("Branching: {} from {} -> {}".format(newbranch, oldbranch,
                     self.lookup(oldbranch)))
        if newbranch is None:
            raise Exception("Nothing to create.")
        if oldbranch is None:
            if not self._repository.head_is_unborn:
                oldbranch = self._repository.head
            else:
                raise Exception("No branch to branch from.")
        if not oldbranch.startswith("refs/heads/"):
            oldbranch = "refs/heads/" + oldbranch
        if not newbranch.startswith("refs/heads/"):
            newbranch = "refs/heads/" + newbranch
        try:
            logger.debug("Branching: create reference {} to {}".format(newbranch, oldbranch))
            self._repository.create_reference(newbranch, self.lookup(oldbranch))
        except Exception as e:
            logger.error(e)
            raise e

    def revert(self, reference='', target='', branch=''):
        """Revert a commit."""
        raise Exception('Not yet supported')


class Revision(object):

    def __init__(self, repository, commit):
        self.id = commit.hex
        self.short_id = self.id[:10]
        self.author = commit.author
        self.committer = commit.committer

        self._repository = repository
        self._commit = commit
        self._parents = None
        self._parsed_message = None

    def _extract(self, message):
        """Extract the value from the commit message untill the end of the quote is found."""
        def eatRest(rest):
            escaped = False
            quoted = True
            value = ""
            for char in rest:
                if not escaped:
                    if char == "\\":
                        escaped = True
                        continue
                    elif char == "\"":
                        quoted = False
                        break
                escaped = False
                value += char
            if quoted:
                value += "\n"
            return value, quoted

        captures = {}
        multiLine = False
        key = None
        value = ""
        messageText = ""
        for line in message.split("\n"):
            if not multiLine:
                keyRest = re.search("^(?P<key>[\\w\\d_]+):(?P<rest>.*)", line)
                if not keyRest:
                    messageText += line + "\n"
                else:
                    key = keyRest.group("key")
                    rest = keyRest.group("rest").lstrip()
                    value = ""
                    if rest[0] == "\"":
                        digest, quoted = eatRest(rest[1:])
                        value += digest
                        if quoted:
                            multiLine = True
                        else:
                            captures[key] = value
                    else:
                        value = rest.strip()
                        captures[key] = value
            else:
                digest, quoted = eatRest(line)
                value += digest
                if not quoted:
                    multiLine = False
                    captures[key] = value

        return captures, messageText

    @property
    def properties(self):
        if not self._parsed_message:
            self._parsed_message = self._extract(self._commit.message)
        return self._parsed_message[0]

    @property
    def message(self):
        if not self._parsed_message:
            self._parsed_message = self._extract(self._commit.message)
        return self._parsed_message[1]

    @property
    def parents(self):
        if self._parents is None:
            self._parents = [Revision(self._repository, id)
                             for id in self._commit.parents]
        return self._parents

    def node(self, path=None):
        return Node(self._repository, self._commit, path)


class Node(object):

    DIRECTORY = "dir"
    FILE = "file"

    def __init__(self, repository, commit, path=None):

        if path in (None, '', '.'):
            self.obj = commit.tree
            self.name = ''
            self.kind = Node.DIRECTORY
            self.tree = self.obj
        else:
            try:
                entry = commit.tree[path]
            except KeyError:
                raise NodeNotFound(path, commit.hex)
            self.obj = repository._repository.get(entry.oid)
            self.name = os.path.normpath(path)
            if self.obj.type == pygit2.GIT_OBJ_TREE:
                self.kind = Node.DIRECTORY
                self.tree = self.obj
            elif self.obj.type == pygit2.GIT_OBJ_BLOB:
                self.kind = Node.FILE
                self.blob = self.obj

        self._repository = repository
        self._commit = commit

    @property
    def oid(self):
        return self.obj.id

    @property
    def is_dir(self):
        return self.kind == Node.DIRECTORY

    @property
    def is_file(self):
        return self.kind == Node.FILE

    @property
    def dirname(self):
        return os.path.dirname(self.name)

    @property
    def basename(self):
        return os.path.basename(self.name)

    @property
    def content(self):
        if not self.is_file:
            return None
        return self.blob.data.decode("utf-8")

    def entries(self, recursive=False):
        if isinstance(self.obj, pygit2.Tree):
            for entry in self.obj:
                dirname = self.is_dir and self.name or self.dirname
                node = Node(
                    self._repository, self._commit, '/'.join(
                        x for x in [dirname, entry.name] if x)
                )

                yield node
                if recursive and node.is_dir and node.obj is not None:
                    for x in node.entries(recursive=True):
                        yield x

    @property
    def content_length(self):
        if self.is_file:
            return self.blob.size
        return None

    def history(self):
        walker = self._repository._repository.walk(self._commit.oid, pygit2.GIT_SORT_TIME)

        c0 = self._commit
        e0 = c0.tree[self.name]

        for c1 in walker:
            try:
                e1 = c1.tree[self.name]
                if e0 and e0.oid != e1.oid:
                    yield Node(self._repository, c1, self.name)
            except KeyError:
                return

            c0 = c1
            e0 = e1


from heapq import heappush, heappop


class Index(object):
    def __init__(self, repository):
        self.repository = repository
        self.revision = None
        self.stash = {}
        self.dirty = False

    def set_revision(self, revision):
        try:
            self.revision = self.repository.revision(revision)
        except RevisionNotFound as e:
            raise IndexError(e)

    def add(self, path, contents, mode=None):
        path = os.path.normpath(path)

        oid = self.repository._repository.create_blob(contents)

        self.stash[path] = (oid, mode or pygit2.GIT_FILEMODE_BLOB)

    def remove(self, path):
        path = os.path.normpath(path)

        self.stash[path] = (None, None)

    def commit(self, message, author_name, author_email, **kwargs):
        if self.dirty:
            raise IndexError('Index already commited')

        ref = kwargs.pop('ref', 'HEAD')
        commiter_name = kwargs.pop('commiter_name', author_name)
        commiter_email = kwargs.pop('commiter_email', author_email)
        parents = kwargs.pop(
            'parents', [self.revision.id] if self.revision else [])

        author = pygit2.Signature(author_name, author_email)
        commiter = pygit2.Signature(commiter_name, commiter_email)

        # Create tree
        tree = IndexTree(self)

        for path, (oid, mode) in self.stash.items():
            if oid is None:
                tree.remove(path)
            else:
                tree.add(path, oid, mode)

        oid = tree.write()
        self.dirty = True

        branch = re.sub("refs/heads/", "", ref)
        if not self.repository.is_bare and (branch == self.repository.current_head or
           self.repository.current_head is None):
            try:
                tree = self.repository._repository.get(oid)
                self.repository._repository.checkout_tree(tree)
            except pygit2.GitError as e:
                logger.info("Local changes in working directory of currently checked out branch: "
                            "{}, {}".format(branch, e))
                pass

        return self.repository._repository.create_commit(
            ref, author, commiter, message, oid, parents
        )


class IndexHeap(object):
    def __init__(self):
        self._dict = {}
        self._heap = []

    def __len__(self):
        return len(self._dict)

    def get(self, path):
        return self._dict.get(path)

    def __setitem__(self, path, value):
        if path not in self._dict:
            n = -path.count(os.sep) if path else 1
            heappush(self._heap, (n, path))

        self._dict[path] = value

    def popitem(self):
        key = heappop(self._heap)
        path = key[1]
        return path, self._dict.pop(path)


class IndexTree(object):
    def __init__(self, index):
        self.repository = index.repository
        self.revision = index.revision
        self.builders = IndexHeap()
        if self.revision:
            self.builders[''] = (
                None, self.repository._repository.TreeBuilder(
                    self.revision._commit.tree)
            )
        else:
            self.builders[''] = (
                None, self.repository._repository.TreeBuilder())

    def get_builder(self, path):
        parts = path.split(os.path.sep)

        # Create builders if needed
        for i in range(len(parts)):
            _path = os.path.join(*parts[0:i + 1])

            if self.builders.get(_path):
                continue

            args = []
            try:
                if self.revision:
                    node = self.revision.node(_path)
                    if node.is_file:
                        raise IndexError(
                            'Cannot create a tree builder. "{}" is a file'.format(
                                node.name)
                        )
                    args.append(node.obj.oid)
            except NodeNotFound:
                pass

            self.builders[_path] = (os.path.dirname(
                _path), self.repository._repository.TreeBuilder(*args))

        return self.builders.get(path)[1]

    def add(self, path, oid, mode):
        builder = self.get_builder(os.path.dirname(path))
        builder.insert(os.path.basename(path), oid, mode)

    def remove(self, path):
        self.revision.node(path)
        builder = self.get_builder(os.path.dirname(path))
        builder.remove(os.path.basename(path))

    def write(self):
        """Attach and writes all builders and return main builder oid."""
        # Create trees
        while len(self.builders) > 0:
            path, (parent, builder) = self.builders.popitem()
            if parent is not None:
                oid = builder.write()
                builder.clear()
                self.builders.get(parent)[1].insert(
                    os.path.basename(path), oid, pygit2.GIT_FILEMODE_TREE)

        oid = builder.write()
        builder.clear()

        return oid


class QuitRemoteCallbacks (pygit2.RemoteCallbacks):
    """Set a pygit callback for user authentication when acting with remotes."""

    def __init__(self, session=None):
        self.session = session

    def credentials(self, url, username_from_url, allowed_types):
        """
        The callback to return a suitable authentication method.
        it supports GIT_CREDTYPE_SSH_KEY and GIT_CREDTYPE_USERPASS_PLAINTEXT
        GIT_CREDTYPE_SSH_KEY with an ssh agent configured in the env variable SSH_AUTH_SOCK
          or with id_rsa and id_rsa.pub in ~/.ssh (password must be the empty string)
        GIT_CREDTYPE_USERPASS_PLAINTEXT from the env variables GIT_USERNAME and GIT_PASSWORD
        """
        if pygit2.credentials.GIT_CREDTYPE_SSH_KEY & allowed_types:
            if "SSH_AUTH_SOCK" in os.environ:
                # Use ssh agent for authentication
                return pygit2.KeypairFromAgent(username_from_url)
            else:
                ssh = join(expanduser('~'), '.ssh')
                if "QUIT_SSH_KEY_HOME" in os.environ:
                    ssh = os.environ["QUIT_SSH_KEY_HOME"]
                # public key is still needed because:
                # _pygit2.GitError: Failed to authenticate SSH session:
                # Unable to extract public key from private key file:
                # Method unimplemented in libgcrypt backend
                pubkey = join(ssh, 'id_rsa.pub')
                privkey = join(ssh, 'id_rsa')
                # check if ssh key is available in the directory
                if os.path.isfile(pubkey) and os.path.isfile(privkey):
                    return pygit2.Keypair(username_from_url, pubkey, privkey, "")
                else:
                    raise Exception(
                        "No SSH keys could be found, please specify SSH_AUTH_SOCK or add keys to " +
                        "your ~/.ssh/"
                    )
        elif pygit2.credentials.GIT_CREDTYPE_USERPASS_PLAINTEXT & allowed_types:
            if self.session and "OAUTH_TOKEN" in self.session:
                return pygit2.UserPass(self.session["OAUTH_TOKEN"], 'x-oauth-basic')
            elif "GIT_USERNAME" in os.environ and "GIT_PASSWORD" in os.environ:
                return pygit2.UserPass(os.environ["GIT_USERNAME"], os.environ["GIT_PASSWORD"])
            else:
                raise Exception(
                    "Remote requested plaintext username and password authentication but " +
                    "GIT_USERNAME or GIT_PASSWORD are not set."
                )
        else:
            raise Exception("Only unsupported credential types allowed by remote end")

    def push_update_reference(self, refname, message):
        """This callback is called for a push operation.

        In the case, that the remote rejects a push, message will be set.
        Because we can't raise an Exception here, we have to write it to self.push_error, thus it is
        important to check on the callback object if push_error is not None after a push.
        """
        self.push_error = None
        if message:
            self.push_error = QuitGitPushError(
                "The reference \"{}\" could not be pushed. Remote told us: {}".format(
                    refname.decode("utf-8"), message))
            return -1
        pass



import logging

import os
from pygit2 import Repository, GIT_OBJ_BLOB, GIT_OBJ_TREE
from os import walk
from os.path import join, isfile, relpath
from quit.exceptions import MissingConfigurationError, InvalidConfigurationError
from quit.exceptions import UnknownConfigurationError
from quit.helpers import isAbsoluteUri
from rdflib import Graph, ConjunctiveGraph, Literal, Namespace, URIRef, BNode
from rdflib.plugins.parsers import notation3
from rdflib.namespace import RDF, NamespaceManager
from rdflib.util import guess_format
from urllib.parse import quote, urlencode
from rdflib.term import _is_valid_uri
from uritools import urisplit

logger = logging.getLogger('quit.conf')


class Feature:
    """Represents the fetures passed by the `--feature` parameter."""
    Unknown = 0
    Provenance = 1 << 0
    Persistence = 1 << 1
    GarbageCollection = 1 << 2
    All = Provenance | Persistence | GarbageCollection


class QuitStoreConfiguration():
    """A class that provides information about settings, filesystem and git."""

    quit = Namespace('http://quit.aksw.org/vocab/')
    store = URIRef('http://quit.aksw.org/instance/store')

    def __init__(
        self,
        configfile='config.ttl',
        features=None,
        upstream=None,
        targetdir=None,
        namespace=None,
        oauthclientid=None,
        oauthclientsecret=None
    ):
        """Initialize store configuration.

        This method checks if the config file is given and reads the config file.
        If the config file is missing, it will be generated after analyzing the
        file structure.
        """
        logger = logging.getLogger('quit.conf.QuitConfiguration')
        logger.debug('Initializing configuration object.')

        self.features = features
        self.configchanged = False
        self.sysconf = Graph()
        self.upstream = None
        self.namespace = None
        self.oauthclientid = oauthclientid
        self.oauthclientsecret = oauthclientsecret

        self.nsMngrSysconf = NamespaceManager(self.sysconf)
        self.nsMngrSysconf.bind('', self.quit, override=False)

        self.__initstoreconfig(
            namespace=namespace,
            upstream=upstream,
            targetdir=targetdir,
            configfile=configfile)

    def __initstoreconfig(self, namespace, upstream, targetdir, configfile):
        """Initialize store settings."""
        if isAbsoluteUri(namespace):
            self.namespace = namespace
        else:
            raise InvalidConfigurationError(
                "Quit expects an absolute http(s) base namespace, {} is not absolute.".format(
                    namespace))

        if configfile and isfile(configfile):
            try:
                self.sysconf.parse(configfile, format='turtle')
            except notation3.BadSyntax:
                raise InvalidConfigurationError(
                    "Bad syntax. Configuration file could not be parsed. {}".format(configfile)
                )
            except PermissionError:
                raise InvalidConfigurationError(
                    "Configuration file could not be parsed. Permission denied. {}".format(
                        configfile))
            except Exception as e:
                raise UnknownConfigurationError("UnknownConfigurationError: {}".format(e))

            self.configfile = configfile
        else:
            if not targetdir:
                raise InvalidConfigurationError('No target directory for git repo given')

        if targetdir:
            self.setRepoPath(targetdir)

        if upstream:
            self.setUpstream(upstream)

        return

    def hasFeature(self, flags):
        return flags == (self.features & flags)

    def getBindings(self):
        q = """SELECT DISTINCT ?prefix ?namespace WHERE {{
            {{
                ?ns a <{binding}> ;
                    <{predicate_prefix}> ?prefix ;
                    <{predicate_namespace}> ?namespace .
            }}
        }}""".format(
            binding=self.quit['Binding'], predicate_prefix=self.quit['prefix'],
            predicate_namespace=self.quit['namespace']
        )

        result = self.sysconf.query(q)
        return [(row['prefix'], row['namespace']) for row in result]

    def getDefaultBranch(self):
        """Get the default branch for the Git repository from configuration.

        Returns:
            A string containing the branch name.
        """
        for s, p, o in self.sysconf.triples((None, self.quit.defaultBranch, None)):
            return str(o)

        return None

    def getRepoPath(self):
        """Get the path of Git repository from configuration.

        Returns:
            A string containig the path of the git repo.
        """
        for s, p, o in self.sysconf.triples((None, self.quit.pathOfGitRepo, None)):
            return str(o)

    def getUpstream(self):
        """Get the URI of Git remote from configuration."""
        for s, p, o in self.sysconf.triples((None, self.quit.upstream, None)):
            return str(o)

    def setUpstream(self, origin):
        self.sysconf.remove((None, self.quit.upstream, None))
        self.sysconf.add((self.store, self.quit.upstream, Literal(origin)))

        return

    def setRepoPath(self, path):
        self.sysconf.remove((None, self.quit.pathOfGitRepo, None))
        self.sysconf.add((self.store, self.quit.pathOfGitRepo, Literal(path)))

        return


class QuitGraphConfiguration():
    """A class that keeps track of the relation between named graphs and files."""

    quit = Namespace('http://quit.aksw.org/vocab/')

    def __init__(self, repository):
        """Init graph configuration.

        This method checks if the config file is given and reads the config file.
        If the config file is missing, it will be generated after analyzing the
        file structure.
        """
        logger = logging.getLogger('quit.conf.QuitConfiguration')
        logger.debug('Initializing configuration object.')

        self.repository = repository
        self.configfile = None
        self.mode = None
        self.graphconf = None
        self.graphs = {}
        self.files = {}

    def initgraphconfig(self, rev):
        """Initialize graph settings.

        Public method to initalize graph settings. This method will be run only once.
        """
        if self.graphconf is None:
            self.graphconf = Graph()
            self.nsMngrGraphconf = NamespaceManager(self.graphconf)
            self.nsMngrGraphconf.bind('', self.quit, override=False)

        graph_files, config_files, rdf_files = self.get_blobs_from_repository(rev)

        if len(graph_files) == 0 and len(config_files) == 0:
            self.mode = 'graphfiles'
        elif len(graph_files) > 0 and len(config_files) > 0:
            raise InvalidConfigurationError(
                "Conflict. Found graphfiles and QuitStore configuration file.")
        elif len(graph_files) > 0:
            self.mode = 'graphfiles'
            self.__init_graph_conf_with_blobs(graph_files, rev)
        elif len(config_files) == 1:
            self.mode = 'configuration'
            self.__init_graph_conf_from_configuration(config_files[0], rdf_files)
        else:
            raise InvalidConfigurationError(
                "Conflict. Found more than one QuitStore configuration file.")

    def __init_graph_conf_with_blobs(self, files, rev):
        """Init a repository by analyzing all existing files."""
        for file, values in files.items():
            format = values[0]
            graphFileId = values[1]
            try:
                graphuri = URIRef(self.__get_uri_from_graphfile_blob(graphFileId))
                self.addgraph(file=file, graphuri=graphuri, format=format)
            except Exception as e:
                logger.debug(e)
                logger.warning('No valid URI found in *.graph file for {}, skipped'.format(file))

    def __init_graph_conf_from_configuration(self, configfileId, known_blobs):
        """Init graphs with setting from config.ttl."""
        try:
            configfile = self.repository.get(configfileId)
        except Exception as e:
            raise InvalidConfigurationError(
                "Blob for configfile with id {} not found in repository {}".format(configfileId, e))

        content = configfile.read_raw()

        try:
            self.graphconf.parse(data=content, format='turtle')
        except Exception as e:
            raise InvalidConfigurationError(
                "Configfile could not be parsed {} {}".format(configfileId, e)
            )
        nsQuit = 'http://quit.aksw.org/vocab/'
        query = 'SELECT DISTINCT ?graphuri ?filename ?format WHERE { '
        query += '  ?graph a <' + nsQuit + 'Graph> . '
        query += '  ?graph <' + nsQuit + 'graphUri> ?graphuri . '
        query += '  ?graph <' + nsQuit + 'graphFile> ?filename . '
        query += '  OPTIONAL { ?graph <' + nsQuit + 'hasFormat> ?format .} '
        query += '}'
        result = self.graphconf.query(query)

        for row in result:
            filename = str(row['filename'])
            if row['format'] is None:
                format = guess_format(filename)
            else:
                format = str(row['format'])
            if format != 'nt':
                break
            if filename not in known_blobs.keys():
                break

            graphuri = URIRef(str(row['graphuri']))

            # we store which named graph is serialized in which file
            self.graphs[graphuri] = filename
            self.files[filename] = {
                'serialization': format, 'graph': graphuri, 'oid': known_blobs[filename]}

    def __get_uri_from_graphfile_blob(self, oid):
        """Search for a graph uri in graph file and return it.

        Args
        ----
           oid: String oid of a graph file

        Returns
        -------
            graphuri: String with the graph URI

        """
        try:
            blob = self.repository.get(oid)
        except ValueError:
            logger.debug("Object with OID {} not found in repository.".format(oid))
            return

        content = blob.read_raw().decode().strip()

        if content and _is_valid_uri(content):
            return content
        raise InvalidConfigurationError("No graph URI found in blob with OID {}.".format(oid))

    def addgraph(self, graphuri, file, format=None):
        graphuri_obj = URIRef(graphuri)
        if graphuri_obj in self.graphs.keys():
            return

        self.graphconf.add((self.quit[quote(graphuri)], RDF.type, self.quit.Graph))
        self.graphconf.add((self.quit[quote(graphuri)], self.quit.graphUri, URIRef(graphuri)))
        self.graphconf.add((self.quit[quote(graphuri)], self.quit.graphFile, Literal(file)))
        self.graphs[graphuri_obj] = file

        if format is not None:
            self.graphconf.add((self.quit[quote(graphuri)], self.quit.hasFormat, Literal(format)))
            self.files[file] = {'serialization': format, 'graph': graphuri_obj, 'oid': file}
        else:
            self.files[file] = {'graph': graphuri_obj, 'oid': file}

    def removegraph(self, graphuri):
        self.graphconf.remove((self.quit[quote(graphuri)], None, None))

        if not isinstance(graphuri, URIRef):
            graphuri = URIRef(graphuri)
        if graphuri in self.graphs.keys():
            filename = self.graphs[graphuri]
            del self.files[filename]
            del self.graphs[graphuri]

        return

    def getgraphs(self):
        """Get all graphs known to conf.

        Returns
        -------
            A list containig all graph uris as string,

        """
        return self.graphs

    def getfiles(self):
        """Get all files known to conf.

        Returns
        -------
            A list containig all files as string,

        """
        return self.files

    def getfileforgraphuri(self, graphuri):
        """Get the file for a given graph uri.

        Args
        ----
            graphuri: A String of the named graph

        Returns
        -------
            A string of the path to the file asociated with named graph

        """
        if isinstance(graphuri, str):
            graphuri = URIRef(graphuri)

        if graphuri in self.graphs.keys():
            return self.graphs[graphuri]

        return

    def getgraphurifilemap(self):
        """Get the dictionary of graphuris and their files.

        Returns
        -------
            A dictionary of graphuris and information about their representation in repository.

        """
        return self.graphs

    def getserializationoffile(self, file):
        """Get the file for a given graph uri.

        Args
        ----
            file: A String of a file path

        Returns
        -------
            A string containing the RDF serialization of file

        """
        if file in self.files.keys():
            return self.files[file]['serialization']

        return

    def getgraphuriforfile(self, file):
        """Get the file for a given graph uri.

        Args
        ----
            file: A String of a file path

        Returns
        -------
            A set containing strings of graph uris asociated to that file

        """
        if file in self.files:
            return self.files[file]['graph']

    def get_blobs_from_repository(self, rev):
        """Analyze all blobs of a revision.

        Returns
        -------
            A triple (dictionary, list, dictionary)
            dict: containg names of rdf-files plus their format and oid of graph file.
            list: containing names of config files.
            dict: containing names rdf files plus format and oid.

        """
        def find_blobs(tree, prefix=''):
            # Collect graph files, rdf files and config files
            for entry in tree:
                if entry.type == GIT_OBJ_BLOB:
                    format = guess_format(entry.name)
                    if format is None and entry.name.endswith('.graph'):
                        graph_file_blobs[join(prefix, entry.name)] = entry.id
                    elif format is not None and format == 'nt':
                        rdf_file_blobs[join(prefix, entry.name)] = (entry.id, format)
                    elif format is not None and entry.name == 'config.ttl':
                        config_files.append(str(entry.id))
                elif entry.type == GIT_OBJ_TREE:
                    tree_obj = self.repository[entry.id]
                    find_blobs(tree_obj, join(prefix, entry.name))

        config_files = []
        graph_files = {}
        graph_file_blobs = {}
        rdf_file_blobs = {}

        try:
            commit = self.repository.revparse_single(rev)
        except Exception:
            return graph_files, config_files, rdf_file_blobs

        find_blobs(commit.tree)

        # collect pairs of rdf files and graph files
        for filename in rdf_file_blobs.keys():
            if filename + '.graph' in graph_file_blobs.keys():
                graph_file_blob_id = graph_file_blobs[filename + '.graph']
                graph_files[filename] = (rdf_file_blobs[filename][1], str(graph_file_blob_id))

        return graph_files, config_files, rdf_file_blobs



"""

Code for carrying out Update Operations

"""
import functools

from rdflib import Graph, Variable, URIRef
from rdflib.term import Node

from rdflib.plugins.sparql.sparql import QueryContext
from rdflib.plugins.sparql.evalutils import _fillTemplate, _join
from rdflib.plugins.sparql.evaluate import evalBGP, evalPart

from collections import defaultdict
from itertools import tee
from quit.exceptions import UnSupportedQuery

def _append(dct, identifier, action, items):
    if items:
        if not isinstance(identifier, Node):
            identifier = URIRef(identifier)
        changes = dct.get(identifier, [])
        changes.append((action, items))
        dct[identifier] = changes


def _graphOrDefault(ctx, g):
    if g == 'DEFAULT':
        return ctx.graph
    else:
        return ctx.dataset.get_context(g)


def _graphAll(ctx, g):
    """
    return a list of graphs
    """
    if g == 'DEFAULT':
        return [ctx.graph]
    elif g == 'NAMED':
        return [c for c in ctx.dataset.contexts()
                if c.identifier != ctx.graph.identifier]
    elif g == 'ALL':
        return list(ctx.dataset.contexts())
    else:
        return [ctx.dataset.get_context(g)]


def evalLoad(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#load
    """
    res = {}
    res["type"] = "LOAD"
    res["source"] = u.iri
    res["delta"] = {}

    if not u.graphiri:
        raise UnSupportedQuery("For load queries we need a iriref for a target graph")

    success = False
    loadedGraph = None
    exceptions = []
    formats = [None, 'n3', 'nt', 'turtle']
    for format in formats:
        loadedGraph = Graph()
        try:
            if not format:
                loadedGraph.load(u.iri)
            else:
                loadedGraph.load(u.iri, format=format)
            success = True
            break
        except Exception as e:
            pass
    if not success:
        raise Exception(
            "Could not load %s as either RDF/XML, N3, Turtle, or NTriples" % (
            u.iri))

    graph = ctx.dataset.get_context(u.graphiri)
    graph += loadedGraph

    _append(res["delta"], u.graphiri, 'additions', loadedGraph)

    return res


def evalCreate(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#create
    """
    g = ctx.datset.get_context(u.graphiri)
    if len(g) > 0:
        raise Exception("Graph %s already exists." % g.identifier)
    raise Exception("Create not implemented!")


def evalClear(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#clear
    """
    for g in _graphAll(ctx, u.graphiri):
        g.remove((None, None, None))


def evalDrop(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#drop
    """
    if ctx.dataset.store.graph_aware:
        for g in _graphAll(ctx, u.graphiri):
            ctx.dataset.store.remove_graph(g)
    else:
        evalClear(ctx, u)


def evalInsertData(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#insertData
    """

    res = {}
    res["type"] = "INSERT"
    res["delta"] = {}

    # add triples
    g = ctx.graph
    filled = list(filter(lambda triple: triple not in g, u.triples))
    if filled:
        _append(res["delta"], 'default', 'additions', filled)
        g += filled

    # add quads
    # u.quads is a dict of graphURI=>[triples]
    for g in u.quads:
        cg = ctx.dataset.get_context(g)
        filledq = list(filter(lambda triple: triple not in cg, u.quads[g]))
        if filledq:
            _append(res["delta"], cg.identifier, 'additions', filledq)
            cg += filledq

    return res


def evalDeleteData(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#deleteData
    """
    res = {}
    res["type"] = "DELETE"
    res["delta"] = {}

    # remove triples
    g = ctx.graph
    filled = list(filter(lambda triple: triple in g, u.triples))
    if filled:
        _append(res["delta"], 'default', 'removals', filled)
        g -= filled

    # remove quads
    # u.quads is a dict of graphURI=>[triples]
    for g in u.quads:
        cg = ctx.dataset.get_context(g)
        filledq = list(filter(lambda triple: triple in cg, u.quads[g]))
        if filledq:
            _append(res["delta"], cg.identifier, 'removals', filledq)
            cg -= filledq

    return res


def evalDeleteWhere(ctx, u):
    """
    http://www.w3.org/TR/sparql11-update/#deleteWhere
    """

    res = {}
    res["type"] = "DELETEWHERE"
    res["delta"] = {}

    _res = evalBGP(ctx, u.triples)
    for g in u.quads:
        cg = ctx.dataset.get_context(g)
        c = ctx.pushGraph(cg)
        _res = _join(_res, list(evalBGP(c, u.quads[g])))

    for c in _res:
        g = ctx.graph
        filled, filled_delta = tee(_fillTemplate(u.triples, c))
        _append(res["delta"], 'default', 'removals', list(filled_delta))
        g -= filled

        for g in u.quads:
            cg = ctx.dataset.get_context(c.get(g))
            filledq, filledq_delta = tee(_fillTemplate(u.quads[g], c))
            _append(res["delta"], cg.identifier, 'removals', list(filledq_delta))
            cg -= filledq

    return res


def evalModify(ctx, u):
    originalctx = ctx

    res = {}
    res["type"] = "MODIFY"
    res["delta"] = {}

    # Using replaces the dataset for evaluating the where-clause
    if u.using:
        otherDefault = False

        for d in u.using:
            if d.default:

                if not otherDefault:
                    # replace current default graph
                    dg = Graph()
                    ctx = ctx.pushGraph(dg)
                    otherDefault = True

                ctx.load(d.default, default=True)

            # TODO re-enable original behaviour if USING NAMED works with named graphs
            # https://github.com/AKSW/QuitStore/issues/144
            elif d.named:
                if otherDefault:
                    ctx = originalctx  # restore original default graph
                raise UnSupportedQuery
            #     g = d.named
            #     ctx.load(g, default=False)

    # "The WITH clause provides a convenience for when an operation
    # primarily refers to a single graph. If a graph name is specified
    # in a WITH clause, then - for the purposes of evaluating the
    # WHERE clause - this will define an RDF Dataset containing a
    # default graph with the specified name, but only in the absence
    # of USING or USING NAMED clauses. In the presence of one or more
    # graphs referred to in USING clauses and/or USING NAMED clauses,
    # the WITH clause will be ignored while evaluating the WHERE
    # clause."
    graphName = 'default'
    if not u.using and u.withClause:
        g = ctx.dataset.get_context(u.withClause)
        graphName = str(g.identifier)
        ctx = ctx.pushGraph(g)

    _res = evalPart(ctx, u.where)

    if u.using:
        if otherDefault:
            ctx = originalctx  # restore original default graph
        if u.withClause:
            g = ctx.dataset.get_context(u.withClause)
            graphName = str(g.identifier)
            ctx = ctx.pushGraph(g)

    for c in _res:
        dg = ctx.graph
        if u.delete:
            filled, filled_delta = tee(_fillTemplate(u.delete.triples, c))
            _append(res["delta"], graphName, 'removals', list(filled_delta))
            dg -= filled

            for g, q in u.delete.quads.items():
                cg = ctx.dataset.get_context(c.get(g))
                filledq, filledq_delta = tee(_fillTemplate(q, c))
                _append(res["delta"], cg.identifier, 'removals', list(filledq_delta))
                cg -= filledq

        if u.insert:
            filled, filled_delta = tee(_fillTemplate(u.insert.triples, c))
            _append(res["delta"], graphName, 'additions', list(filled_delta))
            dg += filled

            for g, q in u.insert.quads.items():
                cg = ctx.dataset.get_context(c.get(g))
                filledq, filledq_delta = tee(_fillTemplate(q, c))
                _append(res["delta"], cg.identifier, 'additions', list(filledq_delta))
                cg += filledq

    return res


def evalAdd(ctx, u):
    """

    add all triples from src to dst

    http://www.w3.org/TR/sparql11-update/#add
    """
    src, dst = u.graph

    srcg = _graphOrDefault(ctx, src)
    dstg = _graphOrDefault(ctx, dst)

    if srcg.identifier == dstg.identifier:
        return

    dstg += srcg


def evalMove(ctx, u):
    """
    remove all triples from dst
    add all triples from src to dst
    remove all triples from src

    http://www.w3.org/TR/sparql11-update/#move
    """

    src, dst = u.graph

    srcg = _graphOrDefault(ctx, src)
    dstg = _graphOrDefault(ctx, dst)

    if srcg.identifier == dstg.identifier:
        return

    dstg.remove((None, None, None))

    dstg += srcg

    if ctx.dataset.store.graph_aware:
        ctx.dataset.store.remove_graph(srcg)
    else:
        srcg.remove((None, None, None))


def evalCopy(ctx, u):
    """
    remove all triples from dst
    add all triples from src to dst

    http://www.w3.org/TR/sparql11-update/#copy
    """

    src, dst = u.graph

    srcg = _graphOrDefault(ctx, src)
    dstg = _graphOrDefault(ctx, dst)

    if srcg.identifier == dstg.identifier:
        return

    dstg.remove((None, None, None))

    dstg += srcg


def evalUpdate(graph, update, initBindings=None, actionLog=False):
    """
    http://www.w3.org/TR/sparql11-update/#updateLanguage

    'A request is a sequence of operations [...] Implementations MUST
    ensure that operations of a single request are executed in a
    fashion that guarantees the same effects as executing them in
    lexical order.

    Operations all result either in success or failure.

    If multiple operations are present in a single request, then a
    result of failure from any operation MUST abort the sequence of
    operations, causing the subsequent operations to be ignored.'

    This will return None on success and raise Exceptions on error

    """

    res = []

    for u in update:

        ctx = QueryContext(graph)
        ctx.prologue = u.prologue

        if initBindings:
            for k, v in initBindings.items():
                if not isinstance(k, Variable):
                    k = Variable(k)
                ctx[k] = v

        try:
            if u.name == 'Load':
                result = evalLoad(ctx, u)
                if result:
                    res.append(result)
            elif u.name == 'Clear':
                evalClear(ctx, u)
            elif u.name == 'Drop':
                evalDrop(ctx, u)
            elif u.name == 'Create':
                evalCreate(ctx, u)
            elif u.name == 'Add':
                evalAdd(ctx, u)
            elif u.name == 'Move':
                evalMove(ctx, u)
            elif u.name == 'Copy':
                evalCopy(ctx, u)
            elif u.name == 'InsertData':
                result = evalInsertData(ctx, u)
                if result:
                    res.append(result)
            elif u.name == 'DeleteData':
                result = evalDeleteData(ctx, u)
                if result:
                    res.append(result)
            elif u.name == 'DeleteWhere':
                result = evalDeleteWhere(ctx, u)
                if result:
                    res.append(result)
            elif u.name == 'Modify':
                result = evalModify(ctx, u)
                if result:
                    res.append(result)
            else:
                raise Exception('Unknown update operation: %s' % (u,))
        except UnSupportedQuery as e:
            return res, e
        except Exception:
            if not u.silent:
                raise
    return res, None



#!/usr/bin/env python3
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')))

import logging
from quit.application import getDefaults, parseEnv, parseArgs
from quit.web.app import create_app
from werkzeug.middleware.dispatcher import DispatcherMiddleware

logger = logging.getLogger('quit.run')


def main():
    sys.setrecursionlimit(2 ** 15)

    defaults = getDefaults()
    env = parseEnv()
    args = parseArgs(sys.argv[1:])
    args = {**defaults, **env, **args}
    application = create_app(args)

    # Set the basepath
    if args['basepath']:
        logger.info("Configure DispatcherMiddleware for basepath \"{}\"".format(args['basepath']))

        def simple(env, resp):
            """A simple WSGI application.

            See also: http://werkzeug.pocoo.org/docs/0.14/middlewares/
            """
            resp('200 OK', [('Content-Type', 'text/plain')])

        application.wsgi_app = DispatcherMiddleware(
            simple, {args['basepath']: application.wsgi_app})

    application.run(debug=args['flask_debug'],
                    use_reloader=False,
                    host=args['host'],
                    port=args['port'])


if __name__ == "__main__":
    main()



import os
import urllib
import hashlib
import rdflib
import logging

from functools import wraps

from flask import Flask, render_template as rt, render_template_string as rts, g, current_app
from flask import url_for, redirect, session, request
from flask_cors import CORS

from jinja2 import Environment, contextfilter, Markup

from quit.application import initialize
from quit.conf import Feature as QuitFeature
from quit.core import MemoryStore, Quit
from quit.git import Repository, QuitRemoteCallbacks
import quit.utils as utils

from quit.namespace import QUIT
from quit.web.service import register
from quit.provenance import Blame

logger = logging.getLogger('quit.web.app')

# For import *
__all__ = ('create_app')

BRANCHES_DROPDOWN_TEMPLATE = """
<span class="dropdown branch-select">
    <button class="btn btn-default dropdown-toggle" type="button" data-toggle="dropdown">
        <i class="fa fa-code-fork" aria-hidden="true"></i> {{ current_ref }}
        <span class="caret"></span>
    </button>
    <ul class="dropdown-menu">
        {% if branches %}
            <li class="dropdown-header">Branches</li>
            {% for branch in branches %}
                <li>
                    <a href="{{ url_for(request.endpoint, branch_or_ref=branch) }}">{{ branch }}</a>
                </li>
            {% endfor %}
        {% endif %}
        {% if tags %}
            <li class="divider"></li>
            <li class="dropdown-header">Tags</li>
            {% for tag in tags %}
                <li><a href="{{ url_for(request.endpoint, branch_or_ref=tag) }}">{{ tag }}</a></li>
            {% endfor %}
        {% endif %}
    </ul>
</span>
"""

BRANCHES_SELECT_DROPDOWN_TEMPLATE = """
<select name="{{ name }}" class="form-control branch-select">
    {% if branches %}
    <optgroup label="Branches">
    {% for branch in branches %}
        <option value="{{ branch }}">{{ branch }}</option>
    {% endfor %}
    </optgroup>
    {% endif %}
    {% if tags %}
    <optgroup label="Tags">
    {% for tag in tags %}
        <option value="{{ tag }}">{{ tag }}</option>
    {% endfor %}
    </optgroup>
    {% endif %}
</select>
"""

REMOTES_DROPDOWN_TEMPLATE = """
<div class="form-group">
<label class="control-label" for="remote">
    <i class="fa fa-cloud" aria-hidden="true"></i> Remote:</span>
</label>
<select name="remote" class="form-control branch-select">
    {% for remote in remotes %}
        <option value="{{ remote.name }}"> {{ remote.name }}
        <small>({{ remote.url }})</small></option>
    {% endfor %}
</select>
</div>
"""

env = Environment()


def feature_required(feature):
    def wrapper(f):
        @wraps(f)
        def decorated_view(*args, **kwargs):
            if not current_app.config['quit'].config.hasFeature(feature):
                return render_template("config_error.html"), 404
            return f(*args, **kwargs)
        return decorated_view
    return wrapper


def create_app(arguments):
    """Create a Flask app."""

    app = Flask(
        __name__.split('.')[0], template_folder='web/templates', static_folder='web/static'
    )
    app.secret_key = os.urandom(24)
    register_app(app, arguments)
    register_hook(app)
    register_blueprints(app)
    register_extensions(app)
    register_errorhandlers(app)
    register_template_helpers(app)

    return app


def register_app(app, arguments):
    """Different ways of configurations."""

    config = initialize(arguments)

    garbageCollection = config.hasFeature(QuitFeature.GarbageCollection)
    logger.debug("Has Garbage collection feature?: {}".format(garbageCollection))

    repository = Repository(config.getRepoPath(), create=True, garbageCollection=garbageCollection,
                            callback=QuitRemoteCallbacks(session=session))
    bindings = config.getBindings()

    quit = Quit(config, repository, MemoryStore(bindings))
    quit.syncAll()

    content = quit.store.store.serialize(format='trig').decode()
    logger.debug("Initialize store with following content: {}".format(content))

    app.config['quit'] = quit
    app.config['blame'] = Blame(quit)
    register(QUIT.service, quit.store.store)


def register_extensions(app):
    """Register extensions."""

    cors = CORS()
    cors.init_app(app)


def register_blueprints(app):
    """Register blueprints in views."""

    from quit.web.modules.debug import debug
    from quit.web.modules.endpoint import endpoint
    from quit.web.modules.git import git
    from quit.web.modules.application import application

    for bp in [debug, endpoint, git, application]:
        app.register_blueprint(bp)

    @app.route("/")
    def index():
        return redirect(url_for('git.commits'))


def register_hook(app):
    import time

    @app.before_request
    def before_request():
        g.start = time.time()

    @app.after_request
    def after_request(response):
        diff = time.time() - g.start
        return response


def register_errorhandlers(app):

    @app.errorhandler(404)
    def page_not_found(error):
        return render_template("404.html"), 404


def register_template_helpers(app):

    @app.template_filter('gravatar')
    def gravatar_lookup(email, size=36):
        gravatar_url = "https://www.gravatar.com/avatar/" + \
            hashlib.md5(email.lower().encode()).hexdigest() + "?"
        gravatar_url += urllib.parse.urlencode({'s': str(size)})
        return gravatar_url

    @contextfilter
    @app.template_filter('term_to_string')
    def term_to_string(ctx, t):
        def qname(ctx, t):
            try:
                config = ctx.get('config')
                link = config['quit'].store.store.compute_qname(t, False)
                return u'%s:%s' % (link[0], link[2])
            except Exception as e:
                return t

        if isinstance(t, rdflib.URIRef):
            link = qname(ctx, t)
            return Markup("<a href='%s'>%s</a>" % (t, link))
        elif isinstance(t, rdflib.Literal):
            if t.language:
                return '"%s"@%s' % (t, t.language)
            elif t.datatype:
                return '"%s"^^&lt;%s&gt;' % (t, qname(ctx, t.datatype))
            else:
                return '"%s"' % t
        return t

    @app.context_processor
    def context_processor():
        def render_branches_dropdown(current_ref, branches, tags):
            return rts(BRANCHES_DROPDOWN_TEMPLATE, current_ref=current_ref, branches=branches,
                       tags=tags)

        def render_branches_select_dropdown(name, branches, tags):
            return rts(BRANCHES_SELECT_DROPDOWN_TEMPLATE, name=name, branches=branches, tags=tags)

        def render_remotes_dropdown(available_remotes):
            return rts(REMOTES_DROPDOWN_TEMPLATE, remotes=available_remotes)

        return dict(render_branches_dropdown=render_branches_dropdown,
                    render_remotes_dropdown=render_remotes_dropdown,
                    render_branches_select_dropdown=render_branches_select_dropdown)


def render_template(template_name_or_list, **kwargs):

    quit = current_app.config['quit']

    available_branches = quit.repository.branches
    available_tags = quit.repository.tags
    available_remotes = quit.repository.remotes
    available_refs = quit.repository.references

    context = {
        'available_refs': available_refs,
        'available_branches': available_branches,
        'available_tags': available_tags,
        'available_remotes': available_remotes,
        'git_timestamp': utils.git_timestamp
    }
    context.update(kwargs)

    return rt(template_name_or_list, **context)



import argparse
import sys
import os
from quit.conf import Feature, QuitStoreConfiguration
from quit.exceptions import InvalidConfigurationError
import rdflib.plugins.sparql
from rdflib.plugins.sparql.algebra import SequencePath
from rdflib.plugin import register
from rdflib.serializer import Serializer
from rdflib.query import Processor, UpdateProcessor, ResultSerializer
import logging

werkzeugLogger = logging.getLogger('werkzeug')
werkzeugLogger.setLevel(logging.INFO)

logger = logging.getLogger('quit')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
ch.setFormatter(formatter)


def initialize(args):
    """Build all needed objects.

    Returns:
        A dictionary containing the store object and git repo object.

    """
    if args['verbose']:
        ch.setLevel(logging.ERROR - args['verbose'] * 10)

    logger.addHandler(ch)
    logger.debug("Parsed args: {}".format(args))

    # add the handlers to the logger
    if args['logfile']:
        try:
            fh = logging.FileHandler(args['logfile'])
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
            logger.debug("Logfile: {}".format(args['logfile']))
        except FileNotFoundError:
            logger.error("Logfile not found: {}".format(args['logfile']))
            sys.exit('Exiting quit')
        except PermissionError:
            logger.error("Can not create logfile: {}".format(args['logfile']))
            sys.exit('Exiting quit')

    # from Github: https://github.com/RDFLib/rdflib/issues/617
    # Egregious hack, the SequencePath object doesn't support compare, this implements the __lt__
    # method so that algebra.py works on sorting in SPARQL queries on e.g. rdf:List paths

    def sequencePathCompareLt(self, other):
        return str(self) < str(other)

    def sequencePathCompareGt(self, other):
        return str(self) < str(other)

    setattr(SequencePath, '__lt__', sequencePathCompareLt)
    setattr(SequencePath, '__gt__', sequencePathCompareGt)
    # End egregious hack

    # To get the best behavior, but still we have https://github.com/RDFLib/rdflib/issues/810
    rdflib.plugins.sparql.SPARQL_DEFAULT_GRAPH_UNION = args['defaultgraph_union']

    # To disable web access: https://github.com/RDFLib/rdflib/issues/810
    rdflib.plugins.sparql.SPARQL_LOAD_GRAPHS = False

    register(
        'sparql', Processor,
        'quit.tools.processor', 'SPARQLProcessor')

    register(
        'sparql', UpdateProcessor,
        'quit.tools.processor', 'SPARQLUpdateProcessor')

    register(
        'application/x-turtle', Serializer,
        'rdflib.plugins.serializers.turtle', 'TurtleSerializer')

    register(
        'application/xml', Serializer,
        'rdflib.plugins.serializers.rdfxml', 'XMLSerializer')

    register(
        'application/json', Serializer,
        'rdflib_jsonld.serializer', 'JsonLDSerializer')

    register(
        'application/trig', Serializer,
        'rdflib.plugins.serializers.trig', 'TrigSerializer')

    register(
        'application/xml', ResultSerializer,
        'rdflib.plugins.sparql.results.xmlresults', 'XMLResultSerializer')

    register(
        'application/json', ResultSerializer,
        'rdflib.plugins.sparql.results.jsonresults', 'JSONResultSerializer')

    register(
        'html', ResultSerializer,
        'quit.plugins.serializers.results.htmlresults', 'HTMLResultSerializer')

    register(
        'text/html', ResultSerializer,
        'quit.plugins.serializers.results.htmlresults', 'HTMLResultSerializer')

    register(
        'application/xhtml+xml', ResultSerializer,
        'quit.plugins.serializers.results.htmlresults', 'HTMLResultSerializer')

    register(
        'text/csv', ResultSerializer,
        'rdflib.plugins.sparql.results.csvresults', 'CSVResultSerializer')

    register(
        'application/sparql-results+xml', ResultSerializer,
        'rdflib.plugins.sparql.results.xmlresults', 'XMLResultSerializer')

    register(
        'application/sparql-results+json', ResultSerializer,
        'rdflib.plugins.sparql.results.jsonresults', 'JSONResultSerializer')

    try:
        config = QuitStoreConfiguration(
            configfile=args['configfile'],
            targetdir=args['targetdir'],
            upstream=args['repourl'],
            features=args['features'],
            namespace=args['namespace'],
            oauthclientid=args['oauth_clientid'],
            oauthclientsecret=args['oauth_clientsecret'],
        )
    except InvalidConfigurationError as e:
        logger.error(e)
        sys.exit('Exiting quit')

    # since repo is handled, we can add graphs to config

    logger.info('QuitStore Configuration initialized.')
    logger.debug('Path of Gitrepo: ' + config.getRepoPath())

    return config


class FeaturesAction(argparse.Action):
    """Actions that are executied for the configuration passed with the `--feature` option."""
    CHOICES = {
        'provenance': Feature.Provenance,
        'persistence': Feature.Persistence,
        'garbagecollection': Feature.GarbageCollection
    }

    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            flags = Feature.Unknown
            for value in values:
                if value not in self.CHOICES.keys():
                    message = ("invalid choice: {0!r} (choose from {1})".format(
                        value, ', '.join([repr(action) for action in self.CHOICES.keys()])))
                    raise argparse.ArgumentError(self, message)
                else:
                    flags |= self.CHOICES[value]

            setattr(namespace, self.dest, flags)


def getDefaults():
    return {
        'port': 5000,
        'host': '::',
        'logfile': None,
        'basepath': None,
        'namespace': 'http://quit.instance/',
        'targetdir': None,
        'repourl': None,
        'configfile': "config.ttl",
        'oauth_clientid': None,
        'oauth_clientsecret': None,
        'verbose': 0,
        'flask_debug': False,
        'defaultgraph_union': False,
        'features': 0
    }


def parseEnv():
    """Parse command line arguments.

    Returns:
        parsed object representing the config arguments.
    """

    env = {}
    if 'QUIT_PORT' in os.environ:
        env['port'] = os.environ['QUIT_PORT']

    if 'QUIT_LOGFILE' in os.environ:
        env['logfile'] = os.environ['QUIT_LOGFILE']

    if 'QUIT_BASEPATH' in os.environ:
        env['basepath'] = os.environ['QUIT_BASEPATH']

    if 'QUIT_NAMESPACE' in os.environ:
        env['namespace'] = os.environ['QUIT_NAMESPACE']

    if 'QUIT_TARGETDIR' in os.environ:
        env['targetdir'] = os.environ['QUIT_TARGETDIR']

    if 'QUIT_CONFIGFILE' in os.environ:
        env['configfile'] = os.environ['QUIT_CONFIGFILE']

    if 'QUIT_OAUTH_CLIENT_ID' in os.environ:
        env['oauth_clientid'] = os.environ['QUIT_OAUTH_CLIENT_ID']

    if 'QUIT_OAUTH_SECRET' in os.environ:
        env['oauth_clientsecret'] = os.environ['QUIT_OAUTH_SECRET']

    return env


def parseArgs(args):
    """Parse command line arguments.

    Returns:
        parsed object representing the config arguments.
    """
    basepathhelp = "Base path (aka. application root) (WSGI only)."
    featurehelp = """This option enables additional features of the QuitStore:
                "provenance" - Store provenance information for each revision.
                "persistance" - Store all internal data as rdf graph."""
    confighelp = """Path of config file (turtle). Defaults to ./config.ttl."""
    loghelp = """Path to the log file."""
    targethelp = 'The directory of the local store repository.'
    namespacehelp = """A base namespace that will be applied when dealing with relative URIs in
                    SPARQL UPDATE queries."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--port', type=int)
    parser.add_argument('--host', type=str)
    parser.add_argument('-l', '--logfile', type=str, help=loghelp)
    parser.add_argument('-b', '--basepath', type=str, help=basepathhelp)
    parser.add_argument('-n', '--namespace', type=str, help=namespacehelp)
    parser.add_argument('-t', '--targetdir', type=str, help=targethelp)
    parser.add_argument('-r', '--repourl', type=str, help='A link/URI to a remote repository.')
    parser.add_argument('-c', '--configfile', type=str, help=confighelp)
    parser.add_argument('--oauth-clientid', type=str, dest='oauth_clientid')
    parser.add_argument('--oauth-clientsecret', type=str, dest='oauth_clientsecret')
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('--flask-debug', action='store_true')
    parser.add_argument('--defaultgraph-union', action='store_true')
    parser.add_argument('-f', '--features', nargs='*', action=FeaturesAction,
                        default=Feature.Unknown,
                        help=featurehelp)

    logger.debug("Parsing args: {}".format(args))

    # Remove all values that received the default value None
    parsedArgs = {k: v for k, v in vars(parser.parse_args(args)).items() if v is not None}
    return parsedArgs



import traceback

import logging
from flask import Blueprint, request, current_app, make_response
from rdflib import ConjunctiveGraph
from quit.conf import Feature
from quit import helpers as helpers
from quit.helpers import parse_sparql_request, parse_query_type
from quit.web.app import render_template, feature_required
from quit.exceptions import UnSupportedQuery, SparqlProtocolError, NonAbsoluteBaseError
from quit.exceptions import FromNamedError, QuitMergeConflict, RevisionNotFound
import datetime
import uuid
import base64

logger = logging.getLogger('quit.modules.endpoint')

__all__ = ('endpoint')

endpoint = Blueprint('endpoint', __name__)
resultSetMimetypesDefault = 'application/sparql-results+xml'
askMimetypesDefault = 'application/sparql-results+xml'
rdfMimetypesDefault = 'text/turtle'

resultSetMimetypes = ['application/sparql-results+xml', 'application/xml',
                      'application/sparql-results+json', 'application/json', 'text/csv',
                      'text/html', 'application/xhtml+xml']
askMimetypes = ['application/sparql-results+xml', 'application/xml',
                'application/sparql-results+json', 'application/json', 'text/html',
                'application/xhtml+xml']
rdfMimetypes = ['text/turtle', 'application/x-turtle', 'application/rdf+xml', 'application/xml',
                'application/n-triples', 'application/trig', 'application/ld+json',
                'application/json']


@endpoint.route("/sparql", defaults={'branch_or_ref': None}, methods=['POST', 'GET'])
@endpoint.route("/sparql/<path:branch_or_ref>", methods=['POST', 'GET'])
def sparql(branch_or_ref):
    """Process a SPARQL query (Select or Update).

    Returns:
        HTTP Response with query result: If query was a valid select query.
        HTTP Response 200: If request contained a valid update query.
        HTTP Response 406: If accept header is not acceptable.
    """
    quit = current_app.config['quit']
    default_branch = quit.getDefaultBranch()

    if not branch_or_ref and not quit.repository.is_empty:
        branch_or_ref = default_branch

    logger.debug("Request method: {}".format(request.method))

    query, type, default_graph, named_graph = parse_sparql_request(request)

    if query is None:
        if request.accept_mimetypes.best_match(['text/html']) == 'text/html':
            return render_template('sparql.html', current_ref=branch_or_ref or default_branch,
                                   mode='query')
        else:
            return make_response('No Query was specified or the Content-Type is not set according' +
                                 'to the SPARQL 1.1 standard', 400)
    else:
        # TODO allow USING NAMED when fixed in rdflib
        if len(named_graph) > 0:
            return make_response('FROM NAMED and USING NAMED not supported, yet', 400)

        parse_type = getattr(helpers, 'parse_' + type + '_type')

        try:
            queryType, parsedQuery = parse_type(
                query, quit.config.namespace, default_graph, named_graph)
        except UnSupportedQuery:
            return make_response('Unsupported Query', 400)
        except NonAbsoluteBaseError:
            return make_response('Non absolute Base URI given', 400)
        except SparqlProtocolError:
            return make_response('Sparql Protocol Error', 400)

    if queryType in ['InsertData', 'DeleteData', 'Modify', 'DeleteWhere', 'Load']:
        if branch_or_ref:
            commit_id = quit.repository.revision(branch_or_ref).id
        else:
            commit_id = None

        parent_commit_id = request.values.get('parent_commit_id', None) or None
        if parent_commit_id and parent_commit_id != commit_id:
            resolution_method = request.values.get('resolution_method', None) or None
            if resolution_method == "reject":
                logger.debug("rejecting update because {} is at {} but {} was expected".format(
                             branch_or_ref, commit_id, parent_commit_id))
                return make_response('reject', 409)  # alternative 412
            elif resolution_method in ("merge", "branch"):
                logger.debug(("writing update to a branch of {} because it is at {} but {} was "
                             "expected").format(branch_or_ref, commit_id, parent_commit_id))
                try:
                    quit.repository.lookup(parent_commit_id)
                except RevisionNotFound:
                    return make_response("The provided parent commit (parent_commit_id={}) "
                                         "could not be found.".format(parent_commit_id), 400)

                time = datetime.datetime.now().strftime('%Y-%m-%d-%H%M%S')
                shortUUID = (base64.urlsafe_b64encode(uuid.uuid1().bytes).decode("utf-8")
                             ).rstrip('=\n').replace('/', '_')
                target_branch = "tmp/{}_{}".format(time, shortUUID)
                target_ref = "refs/heads/" + target_branch
                logger.debug("target ref is: {}".format(target_ref))
                oid = quit.applyQueryOnCommit(parsedQuery, parent_commit_id, target_ref,
                                              query=query, default_graph=default_graph,
                                              named_graph=named_graph)

                if resolution_method == "merge":
                    logger.debug(("going to merge update into {} because it is at {} but {} was "
                                 "expected").format(branch_or_ref, commit_id, parent_commit_id))
                    try:
                        quit.repository.merge(target=branch_or_ref, branch=target_ref)
                        oid = quit.repository.revision(branch_or_ref).id
                        # delete temporary branch
                        tmp_branch = quit.repository._repository.branches.get(target_branch)
                        tmp_branch.delete()
                        response = make_response('success', 200)
                        target_branch = branch_or_ref
                    except QuitMergeConflict as e:
                        response = make_response('merge failed', 400)
                else:
                    response = make_response('branched', 200)
                response.headers["X-CurrentBranch"] = target_branch
                response.headers["X-CurrentCommit"] = oid
                return response

                # Add info about temporary branch
        else:
            graph, commitid = quit.instance(parent_commit_id)

            target_head = request.values.get('target_head', branch_or_ref) or default_branch
            target_ref = 'refs/heads/{}'.format(target_head)
            try:
                oid = quit.applyQueryOnCommit(parsedQuery, branch_or_ref, target_ref,
                                              query=query, default_graph=default_graph,
                                              named_graph=named_graph)
                response = make_response('', 200)
                response.headers["X-CurrentBranch"] = target_head
                if oid is not None:
                    response.headers["X-CurrentCommit"] = oid
                else:
                    response.headers["X-CurrentCommit"] = commitid
                return response
            except Exception as e:
                # query ok, but unsupported query type or other problem during commit
                logger.exception(e)
                return make_response('Error after executing the update query.', 400)
    elif queryType in ['SelectQuery', 'DescribeQuery', 'AskQuery', 'ConstructQuery']:
        try:
            graph, commitid = quit.instance(branch_or_ref)
        except Exception as e:
            logger.exception(e)
            return make_response('No branch or reference given.', 400)

        try:
            res = graph.query(parsedQuery)
        except FromNamedError:
            return make_response('FROM NAMED not supported, yet', 400)
        except UnSupportedQuery:
            return make_response('Unsupported Query', 400)

        mimetype = _getBestMatchingMimeType(request, queryType)

        if not mimetype:
            return make_response("Mimetype: {} not acceptable".format(mimetype), 406)

        response = create_result_response(res, mimetype)
        if branch_or_ref:
            response.headers["X-CurrentBranch"] = branch_or_ref
        if commitid:
            response.headers["X-CurrentCommit"] = commitid
        return response
    else:
        logger.debug("Unsupported Type: {}".format(queryType))
        return make_response("Unsupported Query Type: {}".format(queryType), 400)


@endpoint.route("/provenance", methods=['POST', 'GET'])
@feature_required(Feature.Provenance)
def provenance():
    """Process a SPARQL query (Select only) against Provenance endpoint.

    Returns:
        HTTP Response with query result: If query was a valid select query.
        HTTP Response 200: If request contained a valid update query.
        HTTP Response 400: If request doesn't contain a valid sparql query.
        HTTP Response 406: If accept header is not acceptable.
    """
    quit = current_app.config['quit']

    query, type, default_graph, named_graph = parse_sparql_request(request)
    logger.info('Received provenance query: {}'.format(query))

    if query is not None and type == 'query':
        if len(named_graph) > 0:
            return make_response('Unsupported Query, "FROM NAMED not supported, yet"', 400)
        try:
            queryType, parsedQuery = parse_query_type(query)
        except UnSupportedQuery:
            return make_response('Unsupported Query', 400)
        except NonAbsoluteBaseError:
            return make_response('Non absolute Base URI given', 400)
        except SparqlProtocolError:
            return make_response('Sparql Protocol Error', 400)

        graph = quit.store.store

        if queryType not in ['SelectQuery', 'AskQuery', 'ConstructQuery', 'DescribeQuery']:
            return make_response('Unsupported Query Type', 400)

        res = graph.query(query)

        mimetype = _getBestMatchingMimeType(request, queryType)

        if not mimetype:
            return make_response("Mimetype: {} not acceptable".format(mimetype), 406)

        return create_result_response(res, mimetype)
    else:
        if request.accept_mimetypes.best_match(['text/html']) == 'text/html':
            return render_template('sparql.html', mode='provenance')


def _getBestMatchingMimeType(request, queryType):
    if queryType == 'SelectQuery':
        mimetype_default = resultSetMimetypesDefault
        mimetype_list = resultSetMimetypes
    elif queryType == 'AskQuery':
        mimetype_default = askMimetypesDefault
        mimetype_list = askMimetypes
    elif queryType in ['ConstructQuery', 'DescribeQuery']:
        mimetype_default = rdfMimetypesDefault
        mimetype_list = rdfMimetypes

    match_list = [mimetype_default] + mimetype_list
    if 'Accept' in request.headers:
        mimetype = request.accept_mimetypes.best_match(match_list, None)
    else:
        mimetype = mimetype_default

    return mimetype


def create_result_response(res, mimetype):
    """Create a response with the requested serialization."""
    response = make_response(res.serialize(format=mimetype), 200)
    response.headers['Content-Type'] = mimetype
    return response


def edit_store(quit, branch_or_ref, ref, method, args, body, graph):

    def get_where(graph, args):
        s, p, o, c = _spoc(args)

        result = ConjunctiveGraph()

        for subgraph in (
            x for x in graph.store.contexts((s, p, o)) if c is None or x.identifier == c
        ):
            result.addN((s, p, o, subgraph.identifier)
                        for s, p, o in subgraph.triples((None, None, None)))
        return result

    def copy_where(target, graph, args):
        s, p, o, c = _spoc(args)

        for subgraph in (x for x in graph.contexts()):
            target.store.addN((s, p, o, subgraph.identifier)
                              for s, p, o in subgraph.triples((None, None, None)))

    def remove_where(graph, args):
        s, p, o, c = _spoc(args)
        graph.store.remove((s, p, o, c))

    def clear_where(graph, args):
        _, _, _, c = _spoc(args)
        graph.store.remove_context(c)

    def _spoc(args):
        from rdflib import URIRef
        s = args.get('subj', None)
        p = args.get('pred', None)
        o = args.get('obj', None)
        c = args.get('context', None)
        if c:
            c = URIRef(c)
        return s, p, o, c

    try:
        if method in ['GET', 'HEAD']:
            headers = {"Content-type": 'application/n-quads'}
            response = (200, headers, get_where(graph, args).serialize(format='nquads'))

        elif method == 'DELETE':
            remove_where(graph, args)
            response = (200, dict(), None)

            quit.commit(graph, 'New Commit from QuitStore', branch_or_ref, ref)

        elif method in ['POST', 'PUT']:

            data = ConjunctiveGraph()
            data.parse(data=body, format="nquads")

            if method == 'POST':
                copy_where(graph, data, args)
                response = (200, dict(), None)

            elif method == 'PUT':
                clear_where(graph, args)
                copy_where(graph, data, args)
                response = (200, dict(), None)

            quit.commit(graph, 'New Commit from QuitStore', branch_or_ref, ref)

        else:
            response = (405, {"Allow": "GET, HEAD, POST, PUT, DELETE"},
                        "Method %s not supported" % method)

    except Exception as e:
        logger.exception(e)
        response = (400, dict(), "<pre>{}</pre>".format(traceback.format_exc()))

    return response


@endpoint.route(
    "/statements", defaults={'branch_or_ref': None},
    methods=['GET', 'POST', 'PUT', 'DELETE']
)
@endpoint.route("/statements/<branch_or_ref>", methods=['GET', 'POST', 'PUT', 'DELETE'])
def statements(branch_or_ref):

    quit = current_app.config['quit']
    default_branch = quit.config.getDefaultBranch()

    if not branch_or_ref and not quit.repository.is_empty:
        branch_or_ref = default_branch

    ref = request.values.get('ref', None) or default_branch
    ref = 'refs/heads/{}'.format(ref)

    method = request.method
    args = request.args
    body = request.data.decode('utf-8')

    graph, commitid = quit.instance(branch_or_ref)

    result = edit_store(
        quit=quit,
        branch_or_ref=branch_or_ref,
        ref=ref,
        method=method,
        args=args,
        body=body,
        graph=graph
    )
    code, headers, body = result

    response = make_response(body or '', code)
    for k, v in headers.items():
        response.headers[k] = v
    if commitid:
        response.headers["X-CurrentCommit"] = commitid
    return response



import functools
import logging
from itertools import chain
from rdflib import Graph, ConjunctiveGraph, URIRef
from rdflib.graph import ModificationException
from rdflib.graph import Path


class RewriteGraph(Graph):
    def __init__(
        self, store='default', identifier=None, rewritten_identifier=None, namespace_manager=None
    ):
        super().__init__(
            store=store, identifier=rewritten_identifier, namespace_manager=namespace_manager
        )
        self.__graph = Graph(
            store=store, identifier=identifier, namespace_manager=namespace_manager
        )

    def triples(self, triple):
        return self.__graph.triples(triple)

    def add(self, triple_or_quad):
        raise ModificationException()

    def addN(self, triple_or_quad):
        raise ModificationException()

    def remove(self, triple_or_quad):
        raise ModificationException()

    def __iadd__(self, other):
        raise ModificationException()

    def __isub__(self, other):
        raise ModificationException()

    def parse(self, source, publicID=None, format="xml", **args):
        raise ModificationException()

    def __len__(self):
        return len(self.__graph)


def _copyIfNotExists(store, self, other):
    if other and self not in store.contexts(None):
        store.addN((s, p, o, self) for s, p, o in other.triples((None, None, None)))


class CopyOnEditGraph(Graph):
    def __init__(self, template, store='default', identifier=None, namespace_manager=None):
        super().__init__(store, identifier, namespace_manager)

        assert not template or isinstance(
            template, Graph), "template must be graph"
        assert not template or template.identifier == identifier, "identifier must match"

        self._template = template
        self._store = store

    def add(self, triple_or_quad):
        _copyIfNotExists(self._store, self, self._template)
        super().add(triple_or_quad)

    def addN(self, triple_or_quad):
        _copyIfNotExists(self._store, self, self._template)
        super().addN(triple_or_quad)

    def remove(self, triple_or_quad):
        _copyIfNotExists(self._store, self, self._template)
        super().remove(triple_or_quad)

    def triples(self, triple):
        if self not in self._store.contexts(None):
            return self._template.triples(triple) if self._template else (_ for _ in ())
        else:
            return super().triples(triple)

    @property
    def store(self):
        if self not in self._store.contexts(None):
            return self._template.store
        else:
            return super().store

    def unwrap(self):
        return Graph(store=self.store, identifier=self.identifier)

    def __isub__(self, other):
        """Subtract all triples in Graph other from Graph.

        BNode IDs are not changed.
        """
        for triple in other:
            self.remove(triple)
        return self

    def __len__(self):
        if self not in self._store.contexts(None):
            return self._template.__len__()
        else:
            return super().__len__()


class InMemoryAggregatedGraph(ConjunctiveGraph):
    def __init__(self, store='default', identifier=None, graphs=[]):
        super().__init__(store=store, identifier=None)

        if not (isinstance(graphs, list) and all(isinstance(g, Graph) for g in graphs)):
            raise Exception("graphs argument must be a list of Graphs!!")
        self._contexts = graphs

    def __repr__(self):
        return "<{}: {}|{} graphs>".format(
            type(self).__name__,
            len(self.store.contexts()),
            len((c for c in self.graphs() if c not in self.store.contexts()))
        )

    def _graph(self, c):
        if c is None:
            return None
        if not isinstance(c, Graph):
            return self.get_context(c)
        else:
            return self.get_context(c.identifier)

    def contexts(self, triple=None):

        def collect():
            if triple is None or triple is (None, None, None):
                contexts = (context for context in self._contexts)
            else:
                contexts = (context for context in self._contexts
                            if triple in context)

            seen = set()
            for element in chain(self.store.contexts(triple), contexts):
                k = element.identifier
                if k not in seen:
                    seen.add(k)
                    yield element

        return list(collect())

    graphs = contexts

    def triples(self, triple_or_quad, context=None):
        s, p, o, c = self._spoc(triple_or_quad)
        context = self._graph(context or c)

        if isinstance(p, Path):
            for s, o in p.eval(self, s, o):
                yield s, p, o
        else:
            for graph in self.contexts():
                if context is None or graph.identifier == context.identifier:
                    for s, p, o in graph.triples((s, p, o)):
                        yield s, p, o

    def quads(self, triple_or_quad=None):
        s, p, o, c = self._spoc(triple_or_quad)
        context = self._graph(c)

        for graph in self.graphs():
            if context is None or graph.identifier == context.identifier:
                for s1, p1, o1 in graph.triples((s, p, o)):
                    yield (s1, p1, o1, graph)

    def __contains__(self, triple_or_quad):
        (_, _, _, context) = self._spoc(triple_or_quad)
        context = self._graph(context)

        for graph in self.graphs():
            if context is None or graph.identifier == context.identifier:
                if triple_or_quad[:3] in graph:
                    return True
        return False

    def __len__(self):
        return functools.reduce(lambda a, b: a + len(b), self.contexts(None), 0)

    def _get_context(self, identifier):
        """Search for a context and return the Graph if search was sucessfull.

        Args:
            identifier: URIRef or string of a Graph identifier
        Returns:
            Graph if found, else None
        """
        if isinstance(identifier, URIRef):
            return next((x for x in self._contexts if x.identifier == identifier), None)
        else:
            return next((x for x in self._contexts if str(x.identifier) == identifier), None)

    def get_context(self, identifier, quoted=False):
        """Return the requested context/Graph.

        Args:
            identifier: URIRef, Graph or string
        Returns:
            Graph if found, newly created Graph if not found
        """
        if isinstance(identifier, Graph):
            identifier = identifier.identifier

        context = self._get_context(identifier)

        if context is not None:
            return context
        else:
            return Graph(store=self.store, identifier=identifier, namespace_manager=self)


class InMemoryCopyOnEditAggregatedGraph(InMemoryAggregatedGraph):

    def add(self, triple_or_quad):
        s, p, o, c = self._spoc(triple_or_quad, default=True)

        _copyIfNotExists(self.store, c, self._get_context(self.identifier))

        self.store.add((s, p, o), context=c, quoted=False)

    def addN(self, quads):
        def do(g):
            c = self._graph(g)

            _copyIfNotExists(self.store, c, self._get_context(self.identifier))

            return g

        self.store.addN((s, p, o, do(c)) for s, p, o, c in quads)

    def remove(self, triple_or_quad):
        s, p, o, c = self._spoc(triple_or_quad)

        _copyIfNotExists(self.store, c, self._get_context(self.identifier))

        self.store.remove((s, p, o), context=c)

    def get_context(self, identifier, quoted=False):
        if isinstance(identifier, Graph):
            identifier = identifier.identifier
        return CopyOnEditGraph(
            store=self.store, identifier=identifier, namespace_manager=self,
            template=self._get_context(identifier)
        )



from collections import OrderedDict
from sortedcontainers import SortedSet


class Cache:

    def __init__(self, capacity=50):
        self.stack = OrderedDict()
        self.capacity = capacity

    def get(self, key):
        """Get a value from the cache.

        Raises:
            KeyError if no value was found for the given key
        """
        value = self.stack.pop(key)
        self.stack[key] = value
        return value

    def set(self, key, value):
        try:
            self.stack.pop(key)
        except KeyError:
            if len(self.stack) >= self.capacity:
                self.stack.popitem(last=False)
        self.stack[key] = value

    def remove(self, key):
        try:
            return self.stack.pop(key)
        except KeyError:
            return

    def __contains__(self, key):
        return key in self.stack

    def __iter__(self):
        return (c for c in self.stack)

    @property
    def size(self):
        """
        Return the size of the cache
        """
        return len(self.stack)


class FileReference:
    """A class that manages n-triple files.
    This class stores inforamtation about the location of a n-triple file and is
    able to add and delete triples to that file.
    """

    def __init__(self, path, content):
        """Initialize a new FileReference instance.
        Args:
            filelocation: A string of the filepath.
            filecontentinmem: Boolean to decide if local filesystem should be used to
                or if file content should be kept in memory too . (Defaults false)
        Raises:
            ValueError: If no file at the filelocation, or in the given directory + filelocation.
        """

        if isinstance(content, str):
            new = []
            for line in content.splitlines():
                new.append(' '.join(line.split()))
            content = new

        self._path = path
        self._content = SortedSet(content)
        self._modified = False

    @property
    def path(self):
        return self._path

    @property
    def content(self):
        return "\n".join(self._content) + "\n"

    def add(self, data):
        """Add a triple to the file content."""
        self._content.add(data)

    def extend(self, data):
        """Add triples to the file content."""
        self._content.extend(data)

    def remove(self, data):
        """Remove trple from the file content."""
        try:
            self._content.remove(data)
        except KeyError:
            pass



# Parts of this file were taken from https://github.com/RDFLib/rdflib-web
# Copyrights by Gunnar Aastrand Grimnes <http://gromgull.net/>
#
# LICENSE AGREEMENT FOR RDFLIB
# ------------------------------------------------
# Copyright (c) 2002-2017, RDFLib Team
# See CONTRIBUTORS and http://github.com/RDFLib/rdflib
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
#   * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided
# with the distribution.
#
#   * Neither the name of Daniel Krech nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import rdflib

from jinja2 import Environment, contextfilter, Markup
from rdflib.query import ResultSerializer

__all__ = ('HTMLResultSerializer')

namespace_manager = rdflib.Graph().namespace_manager
namespace_manager.bind('xsd', rdflib.XSD)

nm = rdflib.Graph().namespace_manager
nm.bind('xsd', rdflib.XSD)


def qname(ctx, term):
    try:
        if "graph" in ctx:
            label = ctx["graph"].namespace_manager.compute_qname(term, False)
        else:
            label = nm.compute_qname(term, False)
        return u'%s:%s' % (label[0], label[2])
    except Exception as e:
        return term


@contextfilter
def term_to_string(ctx, term):
    if isinstance(term, rdflib.URIRef):
        label = qname(ctx, term)
        return Markup(u"<a href='%s'>%s</a>" % (term, label))
    elif isinstance(term, rdflib.Literal):
        if term.language:
            return '"%s"@%s' % (term, term.language)
        elif term.datatype:
            return '"%s"^^&lt;%s&gt;' % (term, qname(ctx, term.datatype))
        else:
            return '"%s"' % term
    return term


env = Environment()
env.filters["term_to_string"] = term_to_string

RESULT_TEMPLATE = """
<table class="table">
<thead>
    <tr>
    {% for var in result.vars %}
        <th>{{var}}</th>
    {% endfor %}
    </tr>
</thead>
<tbody>
    {% for row in result.bindings %}
        <tr>
        {% for var in result.vars %}
            <td>{{row[var]|term_to_string}}</td>
        {% endfor %}
        </tr>
    {% endfor %}
</tbody>
</table>
"""


class HTMLResultSerializer(ResultSerializer):

    def __init__(self, result):
        ResultSerializer.__init__(self, result)

    def serialize(self, stream, encoding="utf-8"):
        if self.result.type == 'ASK':
            stream.write("<strong>true</strong>".encode(encoding))
            return
        if self.result.type == 'SELECT':
            template = env.from_string(RESULT_TEMPLATE)
            stream.write(template.render(result=self.result))



#!/usr/bin/env python3

import functools as ft

from rdflib import BNode
from quit.namespace import FOAF, PROV, QUIT


class Blame(object):
    """
    Reusable Blame object for web client
    """

    def __init__(self, quit):
        self.quit = quit

    def _generate_values(self, quads):
        result = list()

        for quad in quads:
            (s, p, o, c) = quad

            c.rewrite = True

            # Todo: BNodes in VALUES are not supported by specification? Using UNDEF for now
            _s = 'UNDEF' if isinstance(s, BNode) else s.n3()
            _p = 'UNDEF' if isinstance(p, BNode) else p.n3()
            _o = 'UNDEF' if isinstance(o, BNode) else o.n3()
            _c = 'UNDEF' if isinstance(c, BNode) else c.identifier.n3()

            c.rewrite = False

            result.append((_s, _p, _o, _c))
        return result

    def run(self, quads=None, branch_or_ref='master'):
        """
        Annotated every quad with the respective author

        Args:
                querystring: A string containing a SPARQL ask or select query.
        Returns:
                The SPARQL result set
        """

        commit = self.quit.repository.revision(branch_or_ref)
        g, commitid = self.quit.instance(branch_or_ref)

        quads = [x for x in g.store.quads((None, None, None))]

        if len(quads) == 0:
            return []

        values = self._generate_values(quads)
        values_string = ft.reduce(lambda acc, quad: acc + '( %s %s %s %s )\n' % quad, values, '')

        q = """
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX aksw: <http://aksw.org/>
            PREFIX quit: <http://quit.aksw.org/vocab/>
            PREFIX prov: <http://www.w3.org/ns/prov#>
            PREFIX foaf: <http://xmlns.com/foaf/0.1/>
            SELECT ?s ?p ?o ?context ?hex ?name ?email ?date WHERE {
                ?commit quit:preceedingCommit* ?c .
                ?c      prov:endedAtTime ?date ;
                        prov:qualifiedAssociation ?qa ;
                        quit:updates ?update ;
                        quit:hex ?hex .
                ?qa     prov:agent ?user ;
                        prov:hadRole quit:Author .
                ?user   foaf:mbox ?email ;
                        rdfs:label ?name .
                ?update quit:graph ?context ;
                        quit:additions ?additions .
                GRAPH ?additions {
                    ?s ?p ?o
                }
                FILTER NOT EXISTS {
                    ?y quit:preceedingCommit+ ?z .
                    ?z quit:updates ?update2 .
                    ?update2 quit:graph ?g ;
                        quit:removals ?removals .
                    GRAPH ?removals {
                        ?s ?p ?o
                    }
                }
                VALUES (?s ?p ?o ?context) {
                    %s
                }
            }
            """ % values_string

        return self.quit.store.store.query(
            q,
            initNs={'foaf': FOAF, 'prov': PROV, 'quit': QUIT},
            initBindings={'commit': QUIT['commit-' + commit.id]}
        )



from rdflib.namespace import Namespace, RDF, RDFS, FOAF, DC, VOID, XSD

__all__ = ('RDF', 'RDFS', 'FOAF', 'DC', 'VOID', 'XSD', 'PROV', 'QUIT', 'is_a')

# missing namespaces
PROV = Namespace('http://www.w3.org/ns/prov#')
QUITD = Namespace('http://quit.aksw.org/vocab/')

QUIT = Namespace('http://quit.aksw.org/vocab/')

# simplified properties
is_a = RDF.type


class Vocabulary:
    Activity = PROV['Activity']
    ImportActivity = QUIT['Import']
    TransformActivity = QUIT['Transformation']

    DataSource = QUIT['dataSource']
    Query = QUIT['query']
    Hex = QUIT['hex']

    StartedAtATime = PROV['startedAtTime']
    EndedAtTime = PROV['endedAtTime']
    Comment = RDFS['comment']

    WasAssociatedWith = PROV['wasAssociatedWith']

    Role = PROV['role']

    Agent = PROV['Agent']
    AgentLabel = RDFS.label
    AgentMail = FOAF.mbox

    Association = PROV['Association']
    QualifiedAssociation = PROV['qualifiedAssociation']
    AssociationAgent = PROV['agent']
    AssociationRole = PROV['role']

    PreceedingCommit = QUIT["preceedingCommit"]

    DiffUpdates = QUIT['updates']
    DiffGraph = QUIT['graph']
    DiffAddition = QUIT['addition']
    DiffRemoval = QUIT['removal']

    SpecializationOf = PROV['specializationOf']
    WasGeneratedBy = PROV['wasGeneratedBy']



from ..exceptions import ServiceException

_services = {}


class Service(object):
    """
    A Service object for sparql
    """

    def __init__(self, name, graph):
        self.name = name
        self.graph = graph


def register(name, graph):
    s = Service(name, graph)
    _services[name] = s


def get(name):
    try:
        s = _services[name]
    except KeyError:
        raise ServiceException("No service registered for %s" % name)
    return s.graph



import pygit2

import logging

from copy import copy

from pygit2 import GIT_MERGE_ANALYSIS_UP_TO_DATE
from pygit2 import GIT_MERGE_ANALYSIS_FASTFORWARD
from pygit2 import GIT_MERGE_ANALYSIS_NORMAL
from pygit2 import GIT_SORT_REVERSE, GIT_RESET_HARD, GIT_STATUS_CURRENT

from rdflib import Graph, ConjunctiveGraph, BNode, Literal, URIRef
import re

from quit.conf import Feature, QuitGraphConfiguration
from quit.helpers import applyChangeset
from quit.namespace import RDFS, FOAF, XSD, PROV, QUIT, is_a
from quit.graphs import RewriteGraph, InMemoryAggregatedGraph
from quit.utils import graphdiff, git_timestamp, iri_to_name
from quit.cache import Cache, FileReference

import subprocess

logger = logging.getLogger('quit.core')


class Queryable:
    """A class that represents a querable graph-like object."""

    def __init__(self, **kwargs):
        pass

    def query(self, querystring):
        """Execute a SPARQL select query.

        Args:
            querystring: A string containing a SPARQL ask or select query.
        Returns:
            The SPARQL result set
        """
        pass

    def update(self, querystring):
        """Execute a SPARQL update query and update the store.

        This method executes a SPARQL update query and updates and commits all affected files.

        Args:
            querystring: A string containing a SPARQL upate query.
        """
        pass


class Store(Queryable):
    """A class that combines and syncronieses n-quad files and an in-memory quad store.

    This class contains information about all graphs, their corresponding URIs and
    pathes in the file system. For every Graph (context of Quad-Store) exists a
    FileReference object (n-quad) that enables versioning (with git) and persistence.
    """

    def __init__(self, store):
        """Initialize a new Store instance."""
        self.store = store

        return


class MemoryStore(Store):
    def __init__(self, additional_bindings=list()):
        store = ConjunctiveGraph(identifier='default')
        nsBindings = [('quit', QUIT), ('foaf', FOAF), ('prov', PROV)]

        for prefix, namespace in nsBindings + additional_bindings:
            store.bind(prefix, namespace)

        super().__init__(store=store)


class VirtualGraph(Queryable):
    def __init__(self, store):
        if not isinstance(store, InMemoryAggregatedGraph):
            raise Exception()
        self.store = store

    def query(self, querystring):
        return self.store.query(querystring)

    def update(self, querystring):
        return self.store.update(querystring)


class Quit(object):
    """Quit object which keeps the store syncronised with the repository."""

    gcProcess = None

    def __init__(self, config, repository, store):
        self.config = config
        self.repository = repository
        self.store = store
        self._commits = Cache()
        self._blobs = Cache()
        self._graphconfigs = Cache()

    def _exists(self, cid):
        uri = QUIT['commit-' + cid]
        for _ in self.store.store.quads((uri, None, None, QUIT.default)):
            return True
        return False

    def getDefaultBranch(self):
        """Get the default branch for the Git repository which should be used in the application.

        This will be the default branch as configured, if it is configured or the current HEAD of
        the repository if the HEAD is born. Will default to "master"

        Returns:
            A string containing the branch name.
        """
        config_default_branch = self.config.getDefaultBranch()
        if config_default_branch:
            return config_default_branch
        repository_current_head = self.repository.current_head
        if repository_current_head:
            return repository_current_head
        return "master"

    def rebuild(self):
        for context in self.store.contexts():
            self.store.remove((None, None, None), context)
        self.syncAll()

    def syncAll(self):
        """Synchronize store with repository data."""
        def traverse(commit, seen):
            commits = []
            merges = []

            while True:
                id = commit.id
                if id in seen:
                    break
                seen.add(id)
                if self._exists(id):
                    break
                commits.append(commit)
                parents = commit.parents
                if not parents:
                    break
                commit = parents[0]
                if len(parents) > 1:
                    merges.append((len(commits), parents[1:]))
            for idx, parents in reversed(merges):
                for parent in parents:
                    commits[idx:idx] = traverse(parent, seen)
            return commits

        seen = set()

        for name in self.repository.tags_or_branches:
            initial_commit = self.repository.revision(name)
            commits = traverse(initial_commit, seen)

            while commits:
                commit = commits.pop()
                self.syncSingle(commit)

    def syncSingle(self, commit):
        if not self._exists(commit.id):
            self.changeset(commit)

    def instance(self, reference, force=False):
        """Create and return dataset for a given commit id.

        Args:
            reference: commit id or reference of the commit to retrieve
            force: force to get the dataset from the git repository instead of the internal cache
        Returns:
            Instance of VirtualGraph representing the respective dataset
        """
        default_graphs = []
        commitid = None

        if reference:
            commit = self.repository.revision(reference)
            commitid = commit.id

            for blob in self.getFilesForCommit(commit):
                try:
                    (name, oid) = blob
                    (f, context) = self.getFileReferenceAndContext(blob, commit)
                    internal_identifier = context.identifier + '-' + str(oid)

                    if force or not self.config.hasFeature(Feature.Persistence):
                        g = context
                    else:
                        g = RewriteGraph(
                            self.store.store.store,
                            internal_identifier,
                            context.identifier
                        )
                    default_graphs.append(g)
                except KeyError:
                    pass

        instance = InMemoryAggregatedGraph(
            graphs=default_graphs, identifier='default')

        return VirtualGraph(instance), commitid

    def changeset(self, commit):

        if (
            not self.config.hasFeature(Feature.Persistence)
        ) and (
            not self.config.hasFeature(Feature.Provenance)
        ):
            return

        g = self.store.store

        if self.config.hasFeature(Feature.Provenance):
            role_author_uri = QUIT['Author']
            role_committer_uri = QUIT['Committer']

            g.add((role_author_uri, is_a, PROV['Role']))
            g.add((role_committer_uri, is_a, PROV['Role']))

        # Create the commit
        i1, commitid = self.instance(commit.id, True)

        commit_uri = QUIT['commit-' + commit.id]

        if self.config.hasFeature(Feature.Provenance):
            g.add((commit_uri, is_a, PROV['Activity']))

            if 'Source' in commit.properties.keys():
                g.add((commit_uri, is_a, QUIT['Import']))
                g.add((commit_uri, is_a, PROV['Usage']))
                sources = commit.properties['Source'].strip()
                for source in re.findall("<.*?>", sources):
                    g.add((commit_uri, QUIT['dataSource'], URIRef(source.strip("<>"))))
            if 'Query' in commit.properties.keys():
                g.add((commit_uri, is_a, QUIT['Transformation']))
                g.add((commit_uri, is_a, PROV['Derivation']))
                g.add((commit_uri, QUIT['query'], Literal(
                    commit.properties['Query'].strip())))

            g.add((commit_uri, QUIT['hex'], Literal(commit.id)))
            g.add((commit_uri, PROV['startedAtTime'], Literal(
                git_timestamp(commit.author.time, commit.author.offset),
                datatype=XSD.dateTime)))
            g.add((commit_uri, PROV['endedAtTime'], Literal(
                git_timestamp(commit.committer.time, commit.committer.offset),
                datatype=XSD.dateTime)))
            g.add((commit_uri, RDFS['label'],
                   Literal(commit.message.strip())))

            # Author
            hash = pygit2.hash(commit.author.email).hex
            author_uri = QUIT['user-' + hash]
            g.add((commit_uri, PROV['wasAssociatedWith'], author_uri))

            g.add((author_uri, is_a, PROV['Agent']))
            g.add((author_uri, RDFS.label, Literal(commit.author.name)))
            g.add((author_uri, FOAF.mbox, Literal(commit.author.email)))

            q_author_uri = BNode()
            g.add((commit_uri, PROV['qualifiedAssociation'], q_author_uri))
            g.add((q_author_uri, is_a, PROV['Association']))
            g.add((q_author_uri, PROV['agent'], author_uri))
            g.add((q_author_uri, PROV['hadRole'], role_author_uri))

            if commit.author.name != commit.committer.name:
                # Committer
                hash = pygit2.hash(commit.committer.email).hex
                committer_uri = QUIT['user-' + hash]
                g.add((commit_uri, PROV['wasAssociatedWith'], committer_uri))

                g.add((committer_uri, is_a, PROV['Agent']))
                g.add((committer_uri, RDFS.label, Literal(commit.committer.name)))
                g.add((committer_uri, FOAF.mbox, Literal(commit.committer.email)))

                q_committer_uri = BNode()
                g.add(
                    (commit_uri, PROV['qualifiedAssociation'], q_committer_uri))
                g.add((q_committer_uri, is_a, PROV['Association']))
                g.add((q_committer_uri, PROV['agent'], author_uri))
                g.add((q_committer_uri, PROV['hadRole'], role_committer_uri))
            else:
                g.add((q_author_uri, PROV['hadRole'], role_committer_uri))

            # Parents
            for parent in iter(commit.parents or []):
                parent_uri = QUIT['commit-' + parent.id]
                g.add((commit_uri, QUIT["preceedingCommit"], parent_uri))
                g.add((commit_uri, PROV["wasInformedBy"], parent_uri))

            # Diff
            parent = next(iter(commit.parents or []), None)

            i2, commitid = self.instance(parent.id, True) if parent else (None, None)

            delta = graphdiff(i2.store if i2 else None, i1.store)

            for index, (iri, changesets) in enumerate(delta.items()):
                update_uri = QUIT['update-{}-{}'.format(commit.id, index)]
                g.add((update_uri, QUIT['graph'], iri))
                g.add((commit_uri, QUIT['updates'], update_uri))
                for (op, triples) in changesets:
                    op_uri = QUIT[op + '-' + commit.id]
                    g.add((update_uri, QUIT[op], op_uri))
                    g.addN((s, p, o, op_uri) for s, p, o in triples)

        # Entities
        if commit.id not in self._graphconfigs:
            self.updateGraphConfig(commit.id)

        map = self._graphconfigs.get(commit.id).getgraphurifilemap()

        for entity in commit.node().entries(recursive=True):
            # todo check if file was changed
            if entity.is_file:

                if entity.name not in map.values():
                    continue

                graphUri = self._graphconfigs.get(commit.id).getgraphuriforfile(entity.name)
                blob = (entity.name, entity.oid)

                try:
                    f, context = self.getFileReferenceAndContext(blob, commit)
                except KeyError:
                    graph = Graph(identifier=graphUri)
                    graph.parse(data=entity.content, format='nt')

                    self._blobs.set(
                        blob, (FileReference(entity.name, entity.content), graph)
                    )

                private_uri = QUIT["graph-{}".format(entity.oid)]

                if (
                    self.config.hasFeature(Feature.Provenance) or
                    self.config.hasFeature(Feature.Persistence)
                ):
                    g.add((private_uri, is_a, PROV['Entity']))
                    g.add(
                        (private_uri, PROV['specializationOf'], context.identifier))
                    g.add(
                        (private_uri, PROV['wasGeneratedBy'], commit_uri))
                    g.add((private_uri, PROV['generatedAtTime'], Literal(
                        git_timestamp(commit.author.time, commit.author.offset),
                        datatype=XSD.dateTime)))

                    q_usage = BNode()
                    g.add((private_uri, PROV['qualifiedGeneration'], q_usage))
                    g.add((q_usage, is_a, PROV['Generation']))
                    g.add((q_usage, PROV['activity'], commit_uri))

                    prev = next(entity.history(), None)
                    if prev:
                        prev_uri = QUIT["graph-{}-{}".format(prev.oid, index)]
                        g.add((private_uri, PROV['wasDerivedFrom'], prev_uri))
                        g.add((commit_uri, PROV['used'], prev_uri))

                        q_derivation = BNode()
                        g.add((private_uri, PROV['qualifiedDerivation'], q_derivation))
                        g.add((q_derivation, is_a, PROV['Derivation']))
                        g.add((q_derivation, PROV['entity'], prev_uri))
                        g.add((q_derivation, PROV['hadActivity'], commit_uri))
                if self.config.hasFeature(Feature.Persistence):
                    g.addN((s, p, o, private_uri) for s, p, o
                           in context.triples((None, None, None)))

    def getFilesForCommit(self, commit):
        """Get all entry, oid tupples for a commit.

        On Cache miss this method also updates the commits cache.
        """

        if commit is None:
            return set()

        if commit.id not in self._commits:
            if commit.id not in self._graphconfigs:
                self.updateGraphConfig(commit.id)

            uriFileMap = self._graphconfigs.get(commit.id).getgraphurifilemap()
            blobs = set()

            for entity in commit.node().entries(recursive=True):
                if entity.is_file:
                    if entity.name not in uriFileMap.values():
                        continue
                    blob = (entity.name, entity.oid)
                    blobs.add(blob)
            self._commits.set(commit.id, blobs)
            return blobs
        return self._commits.get(commit.id)

    def getFileReferenceAndContext(self, blob, commit):
        """Get the FileReference and Context for a given blob (name, oid) of a commit.

        On Cache miss this method also updates teh commits cache.
        """
        if commit.id not in self._graphconfigs:
            self.updateGraphConfig(commit.id)

        if blob not in self._blobs:
            (name, oid) = blob
            content = commit.node(path=name).content
            graphUri = self._graphconfigs.get(commit.id).getgraphuriforfile(name)
            graph = Graph(identifier=URIRef(graphUri))
            graph.parse(data=content, format='nt')
            quitWorkingData = (FileReference(name, content), graph)
            self._blobs.set(blob, quitWorkingData)
            return quitWorkingData
        return self._blobs.get(blob)

    def applyQueryOnCommit(self, parsedQuery, parent_commit_ref, target_ref, query=None,
                           default_graph=[], named_graph=[]):
        """Apply an update query on the graph and the git repository."""
        graph, commitid = self.instance(parent_commit_ref)
        resultingChanges, exception = graph.update(parsedQuery)
        if exception:
            # TODO need to revert or invalidate the graph at this point.
            pass
        oid = self.commit(graph, resultingChanges, 'New Commit from QuitStore', parent_commit_ref,
                          target_ref, query=query, default_graph=default_graph,
                          named_graph=named_graph)
        if exception:
            raise exception
        return oid

    def commit(self, graph, delta, message, parent_commit_ref, target_ref, query=None,
               default_graph=[], named_graph=[], **kwargs):
        """Commit changes after applying deltas to the blobs.

        This methods analyzes the delta and applies the changes to the blobs of the repository.
        A commit message is built with help of message and if called from endpoint with query,
        default_graph and named_graph. **kwargs can be used to extend the commit message with
        custom key-value-pairs.

        Args:
            graph: the current graph instance
            delta: delta that will be applied
            message: commit message
            parent_commit_ref: the commit-id of preceeding commit
            target_ref: a ref/branch were the commit will be applied to
            query: the query that lead to the commit
            default_graph: using-graph-uri values from SPARQL protocol
            named_graph: using-named-graph-uri values from SPARQL protocol
        Returns:
            The newly created commits id
        """
        if self._isDeltaEmpty(delta):
            return

        parent_commit_id = None
        parent_commit = None
        blobs = []
        blobs_new = set()

        if parent_commit_ref:
            parent_commit = self.repository.revision(parent_commit_ref)
        if parent_commit:
            parent_commit_id = parent_commit.id
            try:
                blobs = self.getFilesForCommit(parent_commit)
            except KeyError:
                pass
        index = self.repository.index(parent_commit_id)

        if parent_commit_id not in self._graphconfigs:
            self.updateGraphConfig(parent_commit_id)

        graphconfig = self._graphconfigs.get(parent_commit_id)
        known_files = graphconfig.getfiles().keys()

        blobs_new = self._applyKnownGraphs(delta, blobs, parent_commit, index)
        new_contexts = self._applyUnknownGraphs(delta, known_files)
        new_config = copy(graphconfig)

        for identifier, fileReference in new_contexts.items():
            # Add new blobs to repo
            index.add(fileReference.path, fileReference.content)
            if graphconfig.mode == 'graphfiles':
                index.add(fileReference.path + '.graph', identifier + "\n")

            # Update config
            new_config.addgraph(identifier, fileReference.path, 'nt')

            # Update Cache and add new contexts to store
            blob = fileReference.path, index.stash[fileReference.path][0]
            self._blobs.set(blob, (fileReference, graph.store.get_context(identifier)))
            blobs_new.add(blob)
        if graphconfig.mode == 'configuration':
            index.add('config.ttl', new_config.graphconf.serialize(format='turtle').decode())

        message = self._build_message(message, query, delta, default_graph, named_graph, **kwargs)
        author = self.repository._repository.default_signature

        oid = index.commit(message, author.name, author.email, ref=target_ref)

        if self.config.hasFeature(Feature.GarbageCollection):
            self.garbagecollection()

        if oid:
            self._commits.set(oid.hex, blobs_new)
            commit = self.repository.revision(oid.hex)
            self.syncSingle(commit)

        return oid.hex

    def _build_message(self, message, query, result, default_graph, named_graph, **kwargs):
        out = list()
        if message:
            out.append(message)
            out.append('')
        if query:
            out.append('Query: "{}"'.format(query.replace("\\", "\\\\").replace("\"", "\\\"")))

        source = []
        operation_types = []
        for entry in result:
            if "type" in entry:
                operation_types.append(entry["type"])
            if entry["type"] == "LOAD":
                source.append("<{}>".format(entry["source"]))
        if operation_types:
            out.append('OperationTypes: "{}"'.format(",".join(operation_types)))
        if source:
            out.append('Source: "{}"'.format(",".join(source)))
        if isinstance(default_graph, list) and len(default_graph) > 0:
            out.append('using-graph-uri: {}'.format(', '.join(default_graph)))
        if isinstance(named_graph, list) and len(named_graph) > 0:
            out.append('using-named-graph-uri: {}'.format(', '.join(named_graph)))
        for k, v in kwargs.items():
            out.append('{}: "{}"'.format(k, v.replace('"', "\\\"")))
        return "\n".join(out)

    def _applyKnownGraphs(self, delta, blobs, parent_commit, index):
        blobs_new = set()
        for blob in blobs:
            (fileName, oid) = blob
            try:
                file_reference, context = self.getFileReferenceAndContext(blob, parent_commit)
                for entry in delta:
                    changeset = entry['delta'].get(context.identifier, None)

                    if changeset:
                        applyChangeset(file_reference, changeset, context.identifier)
                        del(entry['delta'][context.identifier])

                index.add(file_reference.path, file_reference.content)

                self._blobs.remove(blob)
                blob = fileName, index.stash[file_reference.path][0]
                self._blobs.set(blob, (file_reference, context))
                blobs_new.add(blob)
            except KeyError:
                pass
        return blobs_new

    def _applyUnknownGraphs(self, delta, known_blobs):
        new_contexts = {}
        for entry in delta:
            for identifier, changeset in entry['delta'].items():
                if isinstance(identifier, BNode) or str(identifier) == 'default':
                    continue  # TODO default graph use case

                if identifier not in new_contexts.keys():
                    fileName = iri_to_name(identifier) + '.nt'

                    if fileName in known_blobs:
                        reg = re.compile(re.escape(iri_to_name(identifier)) + "_([0-9]+).nt")
                        #  n ~ numbers (in blobname), b ~ blobname, m ~ match
                        n = [
                            int(m.group(1)) for b in known_blobs for m in [reg.search(b)] if m
                        ] + [0]
                        fileName = '{}_{}.nt'.format(iri_to_name(identifier), max(n)+1)

                    new_contexts[identifier] = FileReference(fileName, '')

                fileReference = new_contexts[identifier]
                applyChangeset(fileReference, changeset, identifier)
        return new_contexts

    def _isDeltaEmpty(self, result):
        for entry in result:
            if "delta" in entry and entry["delta"]:
                return False
        return True

    def garbagecollection(self):
        """Start garbage collection.
        Args:
            commitid: A string cotaining a commitid.
        """
        try:
            # Check if the garbage collection process is still running
            if self.gcProcess is None or self.gcProcess.poll() is not None:
                # Start garbage collection with "--auto" option,
                # which imidietly terminates, if it is not necessary
                self.gcProcess = subprocess.Popen(
                    ["git", "gc", "--auto", "--quiet"], cwd=self.repository.path
                )
            logger.debug('Spawn git garbage collection.')
        except Exception as e:
            logger.debug('Git garbage collection failed to spawn.')
            logger.debug(e)

    def updateGraphConfig(self, commitId):
        """Update the graph configuration for a given commit id."""
        graphconf = QuitGraphConfiguration(self.repository._repository)
        graphconf.initgraphconfig(commitId)
        self._graphconfigs.set(commitId, graphconf)



import sys
import traceback

from werkzeug.http import parse_accept_header
from flask import Blueprint, flash, redirect, request, url_for, current_app, make_response, Markup
from quit.conf import Feature
from quit.web.app import render_template, feature_required

__all__ = ('debug')

debug = Blueprint('debug', __name__)


@debug.route("/blame", defaults={'branch_or_ref': None}, methods=['GET'])
@debug.route("/blame/<branch_or_ref>", methods=['GET'])
@feature_required(Feature.Provenance)
def blame(branch_or_ref):

    quit = current_app.config['quit']

    if not branch_or_ref and not quit.repository.is_empty:
        branch_or_ref = quit.getDefaultBranch()

    blame = current_app.config['blame']

    if 'Accept' in request.headers:
        mimetype = parse_accept_header(request.headers['Accept']).best
    else:
        mimetype = 'application/sparql-results+json'

    try:
        res = blame.run(branch_or_ref=branch_or_ref)

        if mimetype in ['text/html', 'application/xhtml_xml', '*/*']:
            results = [{'commit': quit.repository.revision(
                row['hex']), 'blame': row} for row in res]
            response = make_response(render_template("blame.html", results=results,
                                     current_ref=branch_or_ref))
            response.headers['Content-Type'] = 'text/html'
            return response
        elif mimetype in ['application/json', 'application/sparql-results+json']:
            response = make_response(res.serialize(format='json'), 200)
            response.headers['Content-Type'] = 'application/json'
            return response
        elif mimetype in [
            'application/rdf+xml', 'application/xml', 'application/sparql-results+xml'
        ]:
            response = make_response(res.serialize(format='xml'), 200)
            response.headers['Content-Type'] = 'application/rdf+xml'
            return response
        elif mimetype in ['application/csv', 'text/csv']:
            response = make_response(res.serialize(format='csv'), 200)
            response.headers['Content-Type'] = 'text/csv'
            return response
    except Exception as e:
        current_app.logger.error(e)
        current_app.logger.error(traceback.format_exc())
        return "<pre>" + traceback.format_exc() + "</pre>", 400



"""
Quit Store
----------------

Quads in Git - Distributed Version Control for RDF Knowledge Bases

The Quit Store (stands for Quads in Git) provides a workspace for distributed
collaborative Linked Data knowledge engineering. You are able to read and write
RDF datasets (aka. multiple Named Graphs) through a standard SPARQL 1.1 Query
and Update interface. To collaborate you can create multiple branches of the
dataset and share your repository with your collaborators as you know it from
Git.
"""
from setuptools import setup
import os

setup(
    name='quit-store',
    version='0.24.0',
    url='https://github.com/AKSW/QuitStore',
    license='GPLv3+',
    author='Natanael Arndt, Norman Radtke',
    author_email='arndtn@gmail.com',
    description='Distributed Version Control for RDF Knowledge Bases',
    long_description=__doc__,
    entry_points={
        'console_scripts': ['quit-store=quit.run:main'],
    },
    packages=[
        'quit',
        'quit.plugins',
        'quit.plugins.serializers',
        'quit.plugins.serializers.results',
        'quit.tools',
        'quit.web',
        'quit.web.extras',
        'quit.web.modules'
    ],
    package_data={
        'quit.web': [
            'static/css/*',
            'static/fonts/*',
            'static/js/*',
            'static/octicons/*',
            'static/octicons/svg/*',
            'templates/*'
        ]
    },
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'rdflib>=4.2.2',
        'Flask',
        'Flask-Cors',
        'sortedcontainers',
        'uritools',
        'pygit2>=1.0.0'
    ],
    dependency_links=[
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: System :: Archiving',
        'Topic :: System :: Distributed Computing'
    ]
)



import os
import pygit2
import rdflib
import logging
from quit.exceptions import QuitMergeConflict, QuitBlobMergeConflict
from rdflib.plugins.serializers.nt import _nt_row as _nt

logger = logging.getLogger('quit.merge')


class Merger(object):
    """This class combines all neccessary methods for performing a merge on two graph commits."""

    def __init__(self, quitRepository, repository):
        """A Merger.

        Keyword arguments:
        repository -- the pygit2 Repository instance
        """
        self.quitRepository = quitRepository
        self._repository = repository

    def merge_analysis(self, target=None, branch=None):
        """Analogou reimplementation of Repository.merge_analysis.

        As of libgit2 f1323d9c161aeeada190fd9615a8b5a9fb8a7f3e "/src/merge.c".
        """
        if target == "HEAD" and self._repository.head_is_unborn:
            return pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD | pygit2.GIT_MERGE_ANALYSIS_UNBORN

        target = self.quitRepository.lookup(target)
        branch = self.quitRepository.lookup(branch)

        ancestor_head = self._repository.merge_base(target, branch)
        if ancestor_head is None:
            # TODO there might be more errors
            return pygit2.GIT_MERGE_ANALYSIS_NONE

        # We're up-to-date if we're trying to merge our own common ancestor.
        if branch == ancestor_head:
            return pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE

        # We're fastforwardable if we're our own common ancestor.
        if target == ancestor_head:
            return pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD | pygit2.GIT_MERGE_ANALYSIS_NORMAL

        return pygit2.GIT_MERGE_ANALYSIS_NORMAL

    def merge_three_way_head(self, branch):
        commitid = self.quitRepository.lookup(branch)
        self._repository.merge(commitid)

        if self._repository.index.conflicts is not None:
            for conflict in self._repository.index.conflicts:
                logger.error('Conflicts found in: {}'.format(conflict[0].path))
            raise QuitMergeConflict('Conflicts, ahhhhh!!')

        user = self._repository.default_signature
        tree = self._repository.index.write_tree()
        self._repository.create_commit(
            'HEAD', user, user, 'Merge!', tree,
            [self._repository.head.target, commitid]
        )
        # We need to do this or git CLI will think we are still merging.
        self._repository.state_cleanup()

    def merge_quit_commits(self, target, branch, favour):

        targetOid = self.quitRepository.lookup(target)
        branchOid = self.quitRepository.lookup(branch)
        baseOid = self._repository.merge_base(targetOid, branchOid)

        baseCommit = self._repository.get(baseOid)
        targetCommit = self._repository.get(targetOid)
        branchCommit = self._repository.get(branchOid)

        diff = targetCommit.tree.diff_to_tree(branchCommit.tree)

        baseTree = baseCommit.tree

        mergedTreeBuilder = self._repository.TreeBuilder(targetCommit.tree)
        logger.debug(diff)

        logger.debug(diff.stats)
        logger.debug("Diff has following patches")
        conflicts = {}
        for p in diff:
            logger.debug("A Patch")
            logger.debug(p)
            logger.debug(p.line_stats)
            # logger.debug(p.patch)
            logger.debug(p.delta)
            logger.debug(p.delta.nfiles)
            logger.debug(p.delta.status_char())
            logger.debug(p.delta.status)
            logger.debug(p.delta.new_file.path)
            logger.debug(p.delta.new_file.size)
            logger.debug(p.delta.new_file.id)
            logger.debug(p.delta.old_file.path)
            logger.debug(p.delta.old_file.size)
            logger.debug(p.delta.old_file.id)
            if p.delta.status == pygit2.GIT_DELTA_UNMODIFIED:
                continue
            # TODO support merging subdirectories
            # TODO support mode changes
            # TODO support merge of moved graphs

            if p.delta.status in [pygit2.GIT_DELTA_ADDED, pygit2.GIT_DELTA_DELETED,
                                  pygit2.GIT_DELTA_MODIFIED]:
                if p.delta.old_file.path in baseTree:
                    baseOid = baseTree[p.delta.old_file.path].id
                else:
                    baseOid = pygit2.Oid(hex=pygit2.GIT_OID_HEX_ZERO)
                try:
                    mergedBlob = self._merge_graph_blobs(p.delta.old_file.id, p.delta.new_file.id,
                                                         baseOid, favour=favour)
                    mergedTreeBuilder.insert(p.delta.old_file.path, mergedBlob,
                                             p.delta.old_file.mode)
                except QuitBlobMergeConflict as mergeconflict:
                    conflicts[p.delta.old_file.path] = mergeconflict.getObject()
            else:
                raise Exception("There are operations which I don't know how to merge. yet.")

        if conflicts:
            raise QuitMergeConflict("Unfortunately, we've experienced a merge conflict!", conflicts)

        mergedTreeOid = mergedTreeBuilder.write()
        if target == "HEAD" or self._repository.head.name == target:
            print(target)
            mergedTree = self._repository.get(mergedTreeOid)
            self._repository.checkout_tree(mergedTree)

        # Create commit with our resulting tree
        user = self._repository.default_signature
        commitId = self._repository.create_commit(
            target, user, user, 'Merge graphs with {}!'.format(favour), mergedTreeOid,
            [targetOid, branchOid]
        )

        logger.debug("Created commit {} and forwarded {}".format(commitId, target))
        # We need to do this or git CLI will think we are still merging.
        self._repository.state_cleanup()

    def _merge_graph_blobs(self, graphAOid, graphBOid, graphBaseOid, favour):
        """Merge two commited graphs, with a base, into one merged blob.

        Returns:
        blob
        """
        logger.debug("blobs to merge: {} {} {}".format(graphAOid, graphBOid, graphBaseOid))
        if graphAOid == graphBaseOid:
            return graphBOid
        if graphBOid == graphBaseOid:
            return graphAOid

        if favour == "three-way":
            return self._merge_threeway_graph_blobs(graphAOid, graphBOid, graphBaseOid)
        if favour == "context":
            return self._merge_context_graph_blobs(graphAOid, graphBOid, graphBaseOid)

    def _merge_threeway_graph_blobs(self, graphAOid, graphBOid, graphBaseOid):
        if str(graphAOid) == pygit2.GIT_OID_HEX_ZERO:
            a = set()
        else:
            graphAblob = self._repository[graphAOid].data
            a = set(graphAblob.decode("utf-8").strip().split("\n"))

        if str(graphBOid) == pygit2.GIT_OID_HEX_ZERO:
            b = set()
        else:
            graphBblob = self._repository[graphBOid].data
            b = set(graphBblob.decode("utf-8").strip().split("\n"))

        if graphBaseOid is not None:
            graphBaseblob = self._repository[graphBaseOid].data
            base = set(graphBaseblob.decode("utf-8").strip().split("\n"))
            addA = a - base
            addB = b - base
            intersect = a.intersection(b)
            merged = sorted(intersect.union(addA).union(addB))
        else:
            merged = a.union(b)
        print("\n".join(merged))

        blob = self._repository.create_blob(("\n".join(merged) + "\n").encode("utf-8"))
        return blob

    def _merge_context_graph_blobs(self, graphAOid, graphBOid, graphBaseOid):
        if str(graphAOid) == pygit2.GIT_OID_HEX_ZERO:
            a = set()
        else:
            graphAblob = self._repository[graphAOid].data
            a = set(graphAblob.decode("utf-8").split("\n"))

        if str(graphBOid) == pygit2.GIT_OID_HEX_ZERO:
            b = set()
        else:
            graphBblob = self._repository[graphBOid].data
            b = set(graphBblob.decode("utf-8").split("\n"))

        if graphBaseOid is not None:
            graphBaseblob = self._repository[graphBaseOid].data
            base = set(graphBaseblob.decode("utf-8").split("\n"))
        else:
            base = set()

        logger.debug("base")
        logger.debug(base)
        logger.debug("a")
        logger.debug(a)
        logger.debug("b")
        logger.debug(b)

        addA = a - base
        delA = base - a
        addB = b - base
        delB = base - b

        ok, conflicts = self._merge_context_conflict_detection(addA - addB, delA - delB,
                                                               addB - addA, delB - delA)

        logger.debug("intersect and ok, then merged")
        logger.debug(a.intersection(b))
        logger.debug(ok)
        merged = sorted(a.intersection(b).union(ok))
        logger.debug(merged)
        print(merged)

        if conflicts is not None:
            print("raised")
            raise QuitBlobMergeConflict('Conflicts, ahhhhh!!', merged, conflicts)

        blob = self._repository.create_blob("\n".join(merged).encode("utf-8"))
        return blob

    def _merge_context_conflict_detection(self, addA, delA, addB, delB):

        def conflictSet(graph, conflictingNodes):
            ok = set()
            conflicts = set()
            for triple in graph.triples((None, None, None)):
                if triple[0] in conflictingNodes or triple[2] in conflictingNodes:
                    conflicts.add(_nt(triple).rstrip())
                else:
                    ok.add(_nt(triple).rstrip())
            return ok, conflicts

        graphAddA = rdflib.ConjunctiveGraph()
        graphAddA.parse(data="\n".join(addA), format="nt")
        graphAddB = rdflib.ConjunctiveGraph()
        graphAddB.parse(data="\n".join(addB), format="nt")
        graphDelA = rdflib.ConjunctiveGraph()
        graphDelA.parse(data="\n".join(delA), format="nt")
        graphDelB = rdflib.ConjunctiveGraph()
        graphDelB.parse(data="\n".join(delB), format="nt")

        conflictingNodes = (graphAddA + graphDelA).all_nodes().intersection(
            (graphAddB + graphDelB).all_nodes())
        print(conflictingNodes)
        logger.debug(conflictingNodes)

        conflicts = {}
        ok = set()

        for key, graph in [("addA", graphAddA), ("delA", graphDelA),
                           ("addB", graphAddB), ("delB", graphDelB)]:
            newOK, conflict = conflictSet(graph, conflictingNodes)
            if len(conflict) > 0:
                conflicts[key] = "\n".join(sorted(conflict))
            if key.startswith("add"):
                ok.update(newOK)

        print("list done")

        if conflicts:
            nodes = []
            for node in conflictingNodes:
                logger.debug(node.n3())
                nodes.append(node.n3())
            conflicts["nodes"] = nodes
        print(conflicts)

        print("OK")
        print(ok)

        return sorted(ok), conflicts or None
